#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Seal5 Flow API."""
import os
import sys
import time
import glob
import tarfile
import atexit
from pathlib import Path
from typing import Optional, List, Dict, Tuple, Union


import git
import seal5.logging

from seal5.logging import (
    Logger,
    check_logging_server,
    initialize_logging_server,
    stop_logging_server,
    update_rotary_logger,
)
from seal5.types import Seal5State, PatchStage
from seal5.settings import Seal5Settings, PatchSettings, DEFAULT_SETTINGS, LLVMConfig, LLVMVersion, FileLoggingSettings

from seal5.dependencies import CDSL2LLVMDependency
from seal5 import utils
from seal5.tools import llvm, cdsl2llvm, inject_patches
from seal5.resources.resources import get_patches, get_test_cfg
from seal5.passes import Seal5Pass, PassType, PassScope, PassManager, filter_passes
import seal5.pass_list as passes
from seal5.testgen_utils import collect_generated_test_files
from seal5.build_cache import combine_hashes, hash_arguments, get_patch_id, query_build_cache


logger = Logger("flow")

TRANSFORM_PASS_MAP = [
    # TODO: Global -> Model
    ("convert_models", passes.convert_models, {}),
    ("filter_models", passes.filter_model, {}),
    ("drop_unused", passes.drop_unused, {}),
    ("eliminate_mod_rfs", passes.eliminate_mod_rfs, {}),
    ("eliminate_rd_cmp_zero", passes.eliminate_rd_cmp_zero, {}),
    ("drop_unused2", passes.drop_unused, {}),
    ("inline_functions", passes.inline_functions, {}),
    ("optimize_model", passes.optimize_model, {}),
    ("infer_types", passes.infer_types, {}),
    ("simplify_trivial_slices", passes.simplify_trivial_slices, {}),
    ("explicit_truncations", passes.explicit_truncations, {}),
    # ("process_settings", passes.process_settings, {}),
    # ("write_yaml", passes.write_yaml, {}),
    # ("process_settings2", passes.process_settings, {}),
    ("detect_behavior_constraints", passes.detect_behavior_constraints, {}),
    ("detect_registers", passes.detect_registers, {}),
    ("collect_register_operands", passes.collect_register_operands, {}),
    ("collect_immediate_operands", passes.collect_immediate_operands, {}),
    ("collect_operand_types", passes.collect_operand_types, {}),
    ("process_settings", passes.process_settings, {}),
    ("write_yaml2", passes.write_yaml, {}),
    ("process_settings3", passes.process_settings, {}),
    ("detect_side_effects", passes.detect_side_effects, {}),
    ("detect_inouts", passes.detect_inouts, {}),
    ("detect_imm_leafs", passes.detect_imm_leafs, {}),
    ("detect_calls", passes.detect_calls, {}),
    ("detect_loops", passes.detect_loops, {}),
    ("annotate_opcodes", passes.annotate_opcodes, {}),
    ("check_pattern_support", passes.check_pattern_support, {}),
    ("write_cdsl_full", passes.write_cdsl, {"split": False, "compat": False}),
    # TODO: determine static constraints (xlen,...) -> subtargetvmap
    # detect memory adressing modes
    # self.detect_adressing_modes(verbose)  # TODO
    # detect legal GMIR ops (and map to selectiondag?)
    # self.detect_legal_ops(verbose=verbose)  # TODO
    # extract costs/heuristics
    # self.extract_costs_and_heuristics(verbose)  # TODO
    # ("split_models", passes.split_models, {"by_set": False, "by_instr": False}),
]

GENERATE_PASS_MAP = [
    ("seal5_td", passes.gen_seal5_td, {}),
    # ("model_td", passes.gen_model_td, {}),
    ("set_td", passes.gen_set_td, {}),
    ("riscv_features", passes.gen_riscv_features_patch, {}),
    ("riscv_isa_infos", passes.gen_riscv_isa_info_patch, {}),
    # ("riscv_instr_formats", passes.gen_riscv_instr_formats_patch, {}),
    ("riscv_register_info", passes.gen_riscv_register_info_patch, {}),
    ("riscv_field_types", passes.gen_riscv_field_types_patch, {}),
    ("riscv_instr_info", passes.gen_riscv_instr_info_patch, {}),
    ("riscv_intrinsics", passes.gen_riscv_intrinsics, {}),
    # subtarget_tests
    # register_types
    # operand_types
    # instruction_formats
    # instruction_infos
    # disassembler
    # mc_tests
    # selection_dag_legalizer
    # scalar_costs
    # simd_costs
    # isel_patterns
    # codegen_test
    ("riscv_gisel_legalizer", passes.gen_riscv_gisel_legalizer_patch, {}),
    # TODO: nested pass lists?
    ("pattern_gen", passes.pattern_gen_pass, {}),
]


def lookup_manual_patch(patch: PatchSettings, allow_missing=False):
    """Lookup manual Seal5 patch."""
    patches = get_patches(patch_name=patch.name, target=patch.target, allow_empty=allow_missing)
    if len(patches) == 0:
        # fallback (undefined target)
        patches = get_patches(patch_name=patch.name, target=None, allow_empty=allow_missing)
    if len(patches) == 0:
        raise RuntimeError(f"Manual patch '{patch.name}' not found!")
    assert len(patches) == 1, "Too many matches"
    res = patches[0]
    if patch.target is None:
        target = res.parent.name
        patch.target = str(target)
    if patch.file is None:
        patch.file = str(res)
    return res


def handle_directory(directory: Optional[Path]):
    """Process passed directory."""
    # TODO: handle environment vars
    if directory is None:
        home_dir = os.getenv("SEAL5_HOME")
        if home_dir is None:
            raise RuntimeError("Unable to resolve SEAL5_HOME")
        directory = home_dir
    if not isinstance(directory, Path):
        directory = Path(directory)
    return directory.resolve()


def handle_meta_dir(meta_dir: Optional[Union[str, Path]], directory: Union[str, Path], _name: str):
    """Handle selection of meta directory."""
    # TODO: handle environment vars
    if meta_dir is None:
        meta_dir = os.getenv("SEAL5_META_DIR")
        if meta_dir is None:
            meta_dir = "default"
    if meta_dir == "default":
        if not isinstance(directory, Path):
            assert isinstance(directory, str)
            directory = Path(directory)
        meta_dir = directory / ".seal5"
    elif meta_dir == "user":
        # config_dir = get_seal5_user_config_dir()
        raise NotImplementedError("store meta dirs in .config/seal5/meta")
    if not isinstance(meta_dir, Path):
        assert isinstance(meta_dir, str)
        meta_dir = Path(meta_dir)
    return meta_dir.resolve()


def create_seal5_directories(path: Path, directories: list):
    """Create Seal5 directories."""
    logger.debug("Creating Seal5 directories")
    if not isinstance(path, Path):
        path = Path(path)
    if not path.is_dir():
        raise RuntimeError(f"Not a diretory: {path}")
    for directory in directories:
        (path / directory).mkdir(parents=True, exist_ok=True)


def add_test_cfg(tests_dir: Path):
    """Add LLVM lit testing config."""
    dest = tests_dir / "lit.cfg.py"
    logger.debug("Creating test cfg %s", dest)
    src = get_test_cfg()
    utils.copy(src, dest)


class Seal5Flow:
    """Seal5 Flow."""

    def __init__(
        self, directory: Optional[Path] = None, meta_dir: Optional[Union[str, Path]] = None, name: Optional[str] = None
    ):
        self.directory: Path = handle_directory(directory)
        self.meta_dir: Path = handle_meta_dir(meta_dir, self.directory, name)
        self.name: str = name
        self.state: Seal5State = Seal5State.UNKNOWN
        self.passes: List[Seal5Pass] = []
        self.repo: Optional[git.Repo] = (
            git.Repo(self.directory) if self.directory.is_dir() and utils.is_populated(self.directory) else None
        )
        self.check()
        self.settings: Seal5Settings = Seal5Settings.from_dict({"meta_dir": str(self.meta_dir), **DEFAULT_SETTINGS})
        # self.settings: Seal5Settings = Seal5Settings(directory=self.directory)
        self.settings.directory = str(self.directory)
        if self.settings.settings_file.is_file():
            self.settings = Seal5Settings.from_yaml_file(self.settings.settings_file)
            self.meta_dir = self.settings._meta_dir
        if self.settings.logs_dir.is_dir():
            initialize_logging_server(
                [
                    FileLoggingSettings(
                        filename=self.settings.log_file_path,
                        level=self.settings.logging.file.level,
                        rotate=self.settings.logging.file.rotate,
                        limit=self.settings.logging.file.limit,
                    )
                ],
                self.settings.logging.console.level,
            )
        self.logger = Logger("flow")
        atexit.register(self.close_servers)
        self.name = self.settings.name if name is None else name
        self.name = "default" if self.name is None else self.name
        self.settings.name = self.name
        self.settings.name = self.settings.name if name is None else name
        self.reset_passes()
        self.create_passes()
        # self.init_global_model()

    # def init_global_model(self, name: str = "Seal5"):
    #     dest = self.inputs_dir / f"{name}.seal5model"
    #     args = [
    #         "-",
    #         "-o",
    #         dest,
    #     ]
    #     utils.python(
    #         "-m",
    #         "seal5.frontends.dummy.writer",
    #         *args,
    #         env=env,
    #         print_func=logger.info if verbose else logger.debug,
    #         live=True,
    #     )

    #     self.update_global_model(name=name)

    # def update_global_model(self, name: str = "Seal5"):
    #     self.process_settings(name)

    def reset_passes(self):
        """Reset Seal5 passes."""
        self.passes = []

    def add_pass(self, pass_: Seal5Pass):
        """Add Seal5 pass."""
        pass_names = [p.name for p in self.passes]
        assert pass_.name not in pass_names, f"Duplicate pass name: {pass_.name}"
        self.passes.append(pass_)

    def add_passes(self, pass_list: List[Seal5Pass]):
        """Add multiple Seal5 passes."""
        for pass_ in pass_list:
            self.add_pass(pass_)

    def create_passes(self):
        """Creation of Seal5 pass pipelines."""
        # Transforms
        for pass_name, pass_handler, pass_options in TRANSFORM_PASS_MAP:
            pass_scope = PassScope.MODEL
            self.add_pass(Seal5Pass(pass_name, PassType.TRANSFORM, pass_scope, pass_handler, options=pass_options))

        # Generates
        for pass_name, pass_handler, pass_options in GENERATE_PASS_MAP:
            if pass_name in ["seal5_td", "riscv_gisel_legalizer", "riscv_field_types"]:  # TODO: refactor
                pass_scope = PassScope.GLOBAL
            else:
                pass_scope = PassScope.MODEL
            self.add_pass(Seal5Pass(pass_name, PassType.GENERATE, pass_scope, pass_handler, options=pass_options))

    def check(self):
        """Check/validate Seal5 flow."""

    def initialize(
        self,
        interactive: bool = False,
        clone: bool = False,
        clone_url: Optional[str] = None,
        clone_ref: Optional[str] = None,
        clone_depth: Optional[int] = None,
        progress: bool = False,
        force: bool = False,
        verbose: bool = False,
        ignore_llvm_imm_types: bool = False,
    ):
        """Initialize Seal5 flow."""
        del verbose  # unused
        update_rotary_logger()
        self.logger.info("Initializing Seal5")
        start = time.time()
        metrics = {}
        sha = None
        version_info = None
        if not self.directory.is_dir():
            if clone is False and not utils.ask_user("Clone LLVM repository?", default=False, interactive=interactive):
                self.logger.error("Target directory does not exist! Aborting...")
                sys.exit(1)
            self.logger.info("Cloning LLVM Repository")
            self.repo, sha, version_info = llvm.clone_llvm_repo(
                self.directory,
                clone_url,
                ref=clone_ref,
                label=self.name,
                git_settings=self.settings.git,
                depth=int(clone_depth) if clone_depth is not None else self.settings.llvm.clone_depth,
                progress=progress,
            )
        else:
            if force:
                self.logger.info("Updating LLVM Repository")
                self.repo, sha, version_info = llvm.clone_llvm_repo(
                    self.directory,
                    clone_url,
                    ref=clone_ref,
                    refresh=True,
                    label=self.name,
                    git_settings=self.settings.git,
                    depth=clone_depth or self.settings.llvm.clone_depth,
                )
        if self.meta_dir.is_dir():
            if force is False and not utils.ask_user(
                "Overwrite existing .seal5 diretcory?", default=False, interactive=interactive
            ):
                self.logger.error("Directory %s already exists! Aborting...", self.meta_dir)
                sys.exit(1)
        self.meta_dir.mkdir(exist_ok=True)
        create_seal5_directories(
            self.meta_dir,
            ["deps", "models", "logs", "build", "install", "temp", "inputs", "gen", "patches", "tests", "cache"],
        )
        if not check_logging_server() and self.settings.logs_dir.is_dir():
            initialize_logging_server(
                [
                    FileLoggingSettings(
                        filename=self.settings.log_file_path,
                        level=self.settings.logging.file.level,
                        rotate=self.settings.logging.file.rotate,
                        limit=self.settings.logging.file.limit,
                    )
                ],
                self.settings.logging.console.level,
            )
        add_test_cfg(self.settings.tests_dir)
        if version_info:
            llvm_version = LLVMVersion(**version_info)
            self.settings.llvm.state.version = llvm_version
        if sha:
            self.settings.llvm.state.base_commit = sha
        if not ignore_llvm_imm_types:
            supported_imm_types = llvm.detect_llvm_imm_types(self.directory)
            self.settings.llvm.state.supported_imm_types = list(supported_imm_types)
        self.settings.save()
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"initialize": metrics})
        self.settings.save()
        self.logger.info("Completed initialization of Seal5")

    def setup(
        self,
        interactive: bool = False,
        force: bool = False,
        progress: bool = False,
        verbose: bool = False,
        skip_patterns: bool = False,
    ):
        """Setup Seal5 dependencies."""
        del interactive  # unused
        del verbose  # unused
        self.logger.info("Installing Seal5 dependencies")
        start = time.time()
        metrics = {}
        self.logger.info("Generating default patches")
        gitignore_patch_settings = PatchSettings(
            name="gitignore",
            stage=int(PatchStage.PHASE_0),
            generated=False,
            target="llvm",
            weak=True,
            priority=100,  # This patch has to be applied first!
        )
        self.settings.add_patch(gitignore_patch_settings)
        llvm_version = self.settings.llvm.state.version
        major = llvm_version.major
        inject_markers_patch_settings = PatchSettings(
            name=f"insert_markers_llvm{major}",
            stage=int(PatchStage.PHASE_0),
            generated=False,
            target="llvm",
            weak=True,
            priority=90,  # This patch has to be applied before patterngen patch!
        )
        self.settings.add_patch(inject_markers_patch_settings)
        if not skip_patterns:
            self.logger.info("Cloning CDSL2LLVM")
            # cdsl2llvm_dependency.clone(self.settings.deps_dir / "cdsl2llvm", overwrite=force, depth=1)
            pattern_gen_settings = self.settings.tools.pattern_gen
            kwargs = {}
            if pattern_gen_settings.clone_url is not None:
                kwargs["clone_url"] = pattern_gen_settings.clone_url
            if pattern_gen_settings.ref is not None:
                kwargs["ref"] = pattern_gen_settings.ref
            llvm_version = self.settings.llvm.state.version
            if llvm_version is not None:
                kwargs["llvm_version"] = llvm_version
            cdsl2llvm_dependency = CDSL2LLVMDependency(**kwargs)
            cdsl2llvm_dependency.clone(
                self.settings.deps_dir / "cdsl2llvm",
                overwrite=force,
                depth=pattern_gen_settings.clone_depth,
                sparse=pattern_gen_settings.sparse_checkout,
                progress=progress,
            )
            integrated_pattern_gen = self.settings.tools.pattern_gen.integrated
            if integrated_pattern_gen:
                has_patterngen_patch = any(patch.name == "pattern_gen_support" for patch in self.settings.patches)
                if not has_patterngen_patch:
                    self.logger.info("Adding PatternGen to target LLVM")
                    patch_settings = cdsl2llvm.get_pattern_gen_patches(
                        self.settings.deps_dir / "cdsl2llvm",
                        self.settings.temp_dir,
                    )
                    self.settings.add_patch(patch_settings)
            else:
                self.logger.info("Building PatternGen")
                llvm_config = LLVMConfig(
                    options={
                        "CMAKE_BUILD_TYPE": "Release",
                        "LLVM_BUILD_TOOLS": False,
                        "LLVM_ENABLE_ASSERTIONS": False,
                        "LLVM_OPTIMIZED_TABLEGEN": True,
                        "LLVM_ENABLE_PROJECT": [],
                        "LLVM_TARGETS_TO_BUILD": ["RISCV"],
                    }
                )
                cmake_options = llvm_config.options
                cdsl2llvm.build_pattern_gen(
                    self.settings.deps_dir / "cdsl2llvm",
                    self.settings.deps_dir / "cdsl2llvm" / "llvm" / "build",
                    cmake_options=cmake_options,
                    use_ninja=self.settings.llvm.ninja,
                )
                self.logger.info("Completed build of PatternGen")
                self.logger.info("Building llc")
                cdsl2llvm.build_llc(
                    self.settings.deps_dir / "cdsl2llvm",
                    self.settings.deps_dir / "cdsl2llvm" / "llvm" / "build",
                    cmake_options=cmake_options,
                    use_ninja=self.settings.llvm.ninja,
                )
                self.logger.info("Completed build of llc")
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"setup": metrics})
        self.settings.save()
        self.logger.info("Completed installation of Seal5 dependencies")

    def load_cfg(self, file: Path, overwrite: bool = False):
        """Load YAML cfg."""
        assert file.is_file(), f"File does not exist: {file}"
        new_settings: Seal5Settings = Seal5Settings.from_yaml_file(file)
        self.settings.merge(new_settings, overwrite=overwrite, inplace=True)
        self.settings.save()

    def load_test(self, file: Path, overwrite: bool = True):
        """Load test file."""
        assert file.is_file(), f"File does not exist: {file}"
        filename: str = file.name
        dest = self.settings.tests_dir / filename
        if dest.is_file() and not overwrite:
            raise RuntimeError(f"File {filename} already loaded!")
        utils.copy(file, dest)
        if str(dest) not in self.settings.test.paths:
            self.settings.test.paths.append(str(dest))
            self.settings.save()

    def prepare_environment(self):
        """Prepare Seal5 environment."""
        env = os.environ.copy()
        # env["PYTHONPATH"] = str(self.settings.deps_dir / "M2-ISA-R")
        cdsl2llvm_build_dir = None
        integrated_pattern_gen = self.settings.tools.pattern_gen.integrated
        if integrated_pattern_gen:
            cdsl2llvm_build_dir = self.settings.get_llvm_build_dir(fallback=True, check=False)
        else:
            cdsl2llvm_build_dir = self.settings.deps_dir / "cdsl2llvm" / "llvm" / "build"
        if cdsl2llvm_build_dir.is_dir():
            cdsl2llvm_build_dir = str(cdsl2llvm_build_dir)
            env["CDSL2LLVM_DIR"] = cdsl2llvm_build_dir
        if seal5.logging.SEAL5_INTERNALS_LOGGING_PORT is not None:
            env["SEAL5_INTERNALS_LOGGING_PORT"] = str(seal5.logging.SEAL5_INTERNALS_LOGGING_PORT)
        return env

    def parse_coredsl(self, file, out_dir, verbose: bool = False, log_level: str = "warning"):
        """Parse CDSL file."""
        args = [
            file,
            "-o",
            out_dir,
            "--log",
            log_level,
        ]
        utils.python(
            "-m",
            "seal5.frontends.coredsl2_seal5.parser",
            *args,
            env=self.prepare_environment(),
            print_func=self.logger.info if verbose else self.logger.debug,
            live=verbose,
        )

    def load_cdsl(self, file: Path, verbose: bool = False, overwrite: bool = False):
        """Load CDSL file."""
        assert file.is_file(), f"File does not exist: {file}"
        filename: str = file.name
        dest = self.settings.inputs_dir / filename
        if dest.is_file() and not overwrite:
            raise RuntimeError(f"File {filename} already loaded!")
        # Add file to inputs directory and settings
        utils.copy(file, dest)
        self.settings.inputs.append(filename)
        # Parse CoreDSL file with M2-ISA-R (TODO: Standalone)
        dest = self.settings.models_dir
        self.parse_coredsl(file, dest, verbose=verbose)
        self.settings.save()

    def load(self, files: List[Path], verbose: bool = False, overwrite: bool = False):
        """Load files into Seal5 flow."""
        self.logger.info("Loading Seal5 inputs")
        # Expand glob patterns

        def glob_helper(file):
            res = glob.glob(str(file))
            assert len(res) > 0, f"No files found for pattern: {file}"
            return list(map(Path, res))

        files = sum([glob_helper(file) for file in files], [])
        for file in files:
            self.logger.info("Processing file: %s", file)
            ext = file.suffix
            if ext.lower() in [".yml", ".yaml"]:
                self.load_cfg(file, overwrite=overwrite)
            elif ext.lower() in [".core_desc"]:
                self.load_cdsl(file, verbose=verbose, overwrite=overwrite)
            elif ext.lower() in [".ll", ".c", ".cc", ".cpp", ".s", ".mir", ".gmir"]:
                self.load_test(file, overwrite=overwrite)
            else:
                raise RuntimeError(f"Unsupported input type: {ext}")
        # TODO: only allow single instr set for now and track inputs in settings
        self.logger.info("Completed load of Seal5 inputs")

    def build(self, config=None, target="all", verbose: bool = False, **kwargs):
        """Build Seal5 LLVM."""
        self.logger.info("Building Seal5 LLVM (%s)", target)
        start = time.time()
        metrics = {}
        if config is None:
            config = self.settings.llvm.default_config
        llvm_config = self.settings.llvm.configs.get(config, None)
        assert llvm_config is not None, f"Invalid llvm config: {config}"
        build_dir = self.settings.get_llvm_build_dir(config=config, fallback=True, check=False)
        cmake_options = llvm_config.options
        ccache_settings = self.settings.llvm.ccache
        if kwargs.get("enable_ccache", False):
            ccache_settings.enable = True
        use_build_cache = False
        if kwargs.get("enable_build_cache", False):
            use_build_cache = True
        cached = False
        if use_build_cache:
            build_hash = combine_hashes(
                hash_arguments(), get_patch_id(Path(self.settings.directory), self.settings.llvm.state.base_commit)
            )
            cache_dir = self.settings.cache_dir
            cached = query_build_cache(build_dir, cache_dir)

        if cached:
            assert build_hash is not None
            logger.info("Using cache %s", build_hash)
        else:
            llvm.build_llvm(
                Path(self.settings.directory),
                build_dir,
                cmake_options=cmake_options,
                target=target,
                use_ninja=self.settings.llvm.ninja or kwargs.get("use_ninja", False),
                ccache_settings=ccache_settings,
                verbose=verbose,
            )
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"build": metrics})
        self.settings.save()
        self.logger.info("Completed build of Seal5 LLVM (%s)", target)

    def install(self, dest: Optional[Union[str, Path]] = None, config=None, verbose: bool = False, **kwargs):
        """Install Seal5 LLVM."""
        start = time.time()
        metrics = {}
        if config is None:
            config = self.settings.llvm.default_config
        # TODO: implement compress?
        if dest is None:
            dest = self.settings.install_dir / config
        if not isinstance(dest, Path):
            dest = Path(dest)
        dest.mkdir(exist_ok=True)
        self.logger.info("Installing Seal5 LLVM to: %s", dest)
        llvm_config = self.settings.llvm.configs.get(config, None)
        assert llvm_config is not None, f"Invalid llvm config: {config}"
        cmake_options = llvm_config.options
        ccache_settings = self.settings.llvm.ccache
        if kwargs.get("enable_ccache", False):
            ccache_settings.enable = True
        llvm.build_llvm(
            Path(self.settings.directory),
            self.settings.get_llvm_build_dir(config=config, fallback=True, check=True),
            cmake_options=cmake_options,
            use_ninja=self.settings.llvm.ninja or kwargs.get("use_ninja", False),
            ccache_settings=ccache_settings,
            target=None,
            install=True,
            install_dir=dest,
            verbose=verbose,
        )
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"build": metrics})
        self.settings.save()
        self.logger.info("Completed install of Seal5 LLVM")

    def transform(self, verbose: bool = False, skip: Optional[List[str]] = None, only: Optional[List[str]] = None):
        """Transform Seal5 models."""
        self.logger.info("Tranforming Seal5 models")
        start = time.time()
        metrics = {"passes": []}
        passes_settings = self.settings.passes
        assert passes_settings is not None
        default_skip = passes_settings.skip
        if skip is None and default_skip:
            skip = default_skip
        default_only = passes_settings.only
        if only is None and default_only:
            only = default_only
        # inplace = True
        # if not inplace:
        #     raise NotImplementedError()

        input_models = self.settings.model_names
        transform_passes = filter_passes(self.passes, pass_type=PassType.TRANSFORM)
        with PassManager("transform_passes", transform_passes, skip=skip, only=only) as pm:
            result = pm.run(input_models, settings=self.settings, env=self.prepare_environment(), verbose=verbose)
            if result:
                metrics_ = result.metrics
                if metrics_:
                    metrics["passes"].append({pm.name: metrics_})

        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"transform": metrics})
        self.settings.save()
        self.logger.info("Completed tranformation of Seal5 models")

    def generate(self, verbose: bool = False, skip: Optional[List[str]] = None, only: Optional[List[str]] = None):
        """Generate Seal5 patches."""
        self.logger.info("Generating Seal5 patches")
        start = time.time()
        metrics = {"passes": []}
        generate_passes = filter_passes(self.passes, pass_type=PassType.GENERATE)
        # TODO: User, Global, PerInstr
        input_models = self.settings.model_names
        with PassManager("generate_passes", generate_passes, skip=skip, only=only) as pm:
            result = pm.run(input_models, settings=self.settings, env=self.prepare_environment(), verbose=verbose)
            if result:
                metrics_ = result.metrics
                if metrics_:
                    metrics["passes"].append({pm.name: metrics_})

        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"generate": metrics})
        self.settings.save()
        self.logger.info("Completed generation of Seal5 patches")

    # def collect_patches(self, stage: Optional[PatchStage]):
    #     return ret

    def collect_patches(self, ignore_priorities: bool = False):
        """Collect Seal5 patches."""
        # generated patches
        temp: Dict[Tuple[str, str], PatchSettings] = {}

        # user-defined patches
        patches_settings = self.settings.patches
        for patch_settings in patches_settings:
            if patch_settings.stage is None:
                patch_settings.stage = int(PatchStage.PHASE_0)
                self.logger.warning("Undefined patch stage for patch %s. Defaulting to PHASE_0", patch_settings.name)
                # raise NotImplementedError("Undefined patch stage!")
            if patch_settings.generated:  # not manual
                if patch_settings.target != "llvm":
                    raise NotImplementedError("Only supporting llvm patches so far")
                if patch_settings.file is None:
                    assert patch_settings.index
                    name = patch_settings.name
                    target = patch_settings.target
                    key = (target, name)
                    dest = self.settings.patches_dir / target
                    dest.mkdir(exist_ok=True)
                    patch_settings.to_yaml_file(dest / f"{name}.yml")
                    # assert key not in temp
                    if key in temp:
                        # override
                        self.logger.debug("Overriding existing patch settings")
                        new = temp[key]
                        new.merge(patch_settings, inplace=True)
                        temp[key] = new
                    else:
                        temp[key] = patch_settings
                else:
                    raise NotImplementedError("Only supporting index based patches so far")
            else:
                patch_file = lookup_manual_patch(patch_settings, allow_missing=True)
                # print("patch_file", patch_file)
                target = patch_settings.target
                name = patch_settings.name
                key = (target, name)
                if key in temp:
                    # override
                    self.logger.debug("Overriding existing patch settings")
                    new = temp[key]
                    new.merge(patch_settings, inplace=True)
                    temp[key] = new
                else:
                    temp[key] = patch_settings
                if patch_file:
                    self.logger.debug("Copying custom patch_file %s", patch_file)
                    dest = self.settings.patches_dir / target
                    dest.mkdir(exist_ok=True)
                    # print("dest", dest)
                    utils.copy(patch_file, dest / f"{name}.patch")
                    patch_settings.to_yaml_file(dest / f"{name}.yml")
        ret = {}
        for patch_settings in temp.values():
            if patch_settings.stage not in ret:
                ret[patch_settings.stage] = []
            ret[patch_settings.stage].append(patch_settings)
        if not ignore_priorities:
            ret = {
                # Pythons sort should be stable, which means patched of equal/default prio will
                # be applied in the order thay havee been added
                stage: list(sorted(patches, key=lambda patch: patch.get_priority(), reverse=True))
                for stage, patches in ret.items()
            }
        return ret

    def resolve_patch_file(self, path):
        """Resolve Seal5 patch file."""
        assert path is not None, "Patch path undefined"
        if isinstance(path, str):
            path = Path(path)
        assert isinstance(path, Path)
        ret = path
        if ret.is_file():
            return path.resolve()
        ret = self.settings.patches_dir / path
        if ret.is_file():
            return ret.resolve()
        raise RuntimeError(f"Patch file {path} not found!")

    def apply_patch(self, patch: PatchSettings, force: bool = False):
        """Apply Seal5 patch."""
        name = patch.name
        target = patch.target
        if patch.enable:
            if patch.check_enabled(self.settings):
                self.logger.info("Applying patch '%s' on '%s'", name, target)
            else:
                return
        else:
            self.logger.info("Skipping patch '%s' on '%s'", name, target)
            return
        if patch.index:
            # use inject_extensions_script
            file = self.settings.patches_dir / target / f"{name}.patch"
            # assert not file.is_file(), f"Patch already exists: {file}"
            prefix = self.settings.git.prefix
            comment = patch.comment
            msg = f"{prefix} {comment}"
            llvm_dir = self.directory
            generated_test_files = collect_generated_test_files(patch.index)
            # TODO: maybe move to pass_list?
            if len(generated_test_files) > 0:
                self.settings.test.paths.extend(generated_test_files)
            inject_patches.generate_patch(
                patch.index,
                llvm_dir=llvm_dir,
                out_file=file,
                author=self.settings.git.author,
                mail=self.settings.git.mail,
                msg=msg,
                append=True,
            )

        else:
            file = self.resolve_patch_file(patch.file)
        dest = None
        if target == "llvm":
            dest = self.directory
            repo = git.Repo(dest)
            if not llvm.check_llvm_repo(dest):
                if force:
                    repo.git.reset(".")
                    repo.git.restore(".")
                else:
                    raise RuntimeError("LLVM repository is not clean!")
        if dest is None:
            raise RuntimeError(f"Unsupported patch target: {target}")
        # TODO: check if clean
        repo.git.apply(file)
        author = self.settings.git.author
        mail = self.settings.git.mail
        actor = git.Actor(author, mail)
        prefix = self.settings.git.prefix
        msg = patch.comment
        if not msg:
            msg = f"Apply patch: {patch.name}"
        if prefix:
            msg = prefix + " " + msg
        repo.git.add(A=True)
        repo.index.commit(msg, author=actor)
        patch.applied = True
        self.settings.save()
        # TODO: commit

    def patch(self, verbose: bool = False, stages: List[PatchStage] = None, force: bool = False):
        """Patch Seal5 LLVM."""
        del verbose  # unused
        self.logger.info("Applying Seal5 patches")
        start = time.time()
        metrics = {}
        if stages is None:
            stages = list(map(PatchStage, range(PatchStage.PHASE_5 + 1)))
        assert len(stages) > 0
        patches_per_stage = self.collect_patches()
        stages_metrics = {}
        for stage in stages:
            self.logger.info("Current stage: %s", stage)
            patches = patches_per_stage.get(stage, [])
            # print("patches", patches)
            for patch in patches:
                if patch.applied:
                    # skipping
                    continue
                self.apply_patch(patch, force=force)
            assert self.repo is not None
            tag_name = f"seal5-{self.name}-stage{int(stage)}"
            tag_msg = f"Patched Seal5 LLVM after stage {stage}"
            # author = git_.get_author(self.settings.git)
            # self.repo.create_tag(tag_name, message=tag_msg, force=True, author=author)
            self.repo.create_tag(tag_name, message=tag_msg, force=True)
            base_tag = f"seal5-{self.name}-base"
            if int(stage) > 0:
                prev_tag = f"seal5-{self.name}-stage{int(stage)-1}"
            else:
                prev_tag = base_tag
            n_files_changed, n_insertions, n_deletions = inject_patches.analyze_diff(
                self.repo, cur=tag_name, base=prev_tag
            )
            stage_metrics = {}
            stage_metrics["n_files_changed"] = n_files_changed
            stage_metrics["n_insertions"] = n_insertions
            stage_metrics["n_deletions"] = n_deletions
            stages_metrics[PatchStage(stage).name] = stage_metrics
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        metrics["stages"] = stages_metrics
        self.settings.metrics.append({"patch": metrics})
        self.settings.save()
        self.logger.info("Completed application of Seal5 patches")

    def test(
        self, debug: bool = False, verbose: bool = False, ignore_error: bool = False, config: Optional[str] = None
    ):
        """Test Seal5 LLVM."""
        del debug  # unused
        self.logger.info("Testing Seal5 LLVM")
        start = time.time()
        metrics = {}
        if config is None:
            config = self.settings.llvm.default_config
        test_paths = self.settings.test.paths
        if len(test_paths) == 0:
            self.logger.warning("No test paths have been specified!")
            passed_tests = []
            failing_tests = []
            score = None
        else:
            passed_tests, failing_tests = llvm.test_llvm(
                self.directory / "llvm" / "test",
                self.settings.get_llvm_build_dir(config=config, fallback=True, check=True),
                test_paths,
                verbose=verbose,
            )
            num_tests = len(passed_tests) + len(failing_tests)
            if num_tests == 0:
                self.logger.warning("No tests have been executed!")
                score = None
            else:
                if len(passed_tests) > 0:
                    self.logger.info("%d tests passed: %s", len(passed_tests), ", ".join(passed_tests))
                if len(failing_tests) > 0:
                    self.logger.error("%d tests failed: %s", len(failing_tests), ", ".join(failing_tests))
                    if not ignore_error:
                        raise RuntimeError("Tests failed!")
                score = len(passed_tests) / num_tests
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        metrics["failing"] = failing_tests
        metrics["passed"] = passed_tests
        metrics["score"] = f"{score*100:.2f}%" if score is not None else None
        self.settings.metrics.append({"test": metrics})
        self.settings.save()
        self.logger.info("Completed test of Seal5 LLVM")

    def deploy(self, dest: Path, verbose: bool = False, stage: PatchStage = PatchStage.PHASE_5):
        """Deploy Seal5 LLVM."""
        del verbose  # unused
        assert dest is not None
        # Archive source files
        self.logger.info("Deploying Seal5 LLVM")
        start = time.time()
        metrics = {}
        # TODO: move to different file
        base_tag = f"seal5-{self.name}-base"
        tag_name = f"seal5-{self.name}-stage{int(stage)}"
        n_files_changed, n_insertions, n_deletions = inject_patches.analyze_diff(self.repo, cur=tag_name, base=base_tag)
        self.repo.git.archive(tag_name, "-o", dest)
        end = time.time()
        diff = end - start
        metrics["n_files_changed"] = n_files_changed
        metrics["n_insertions"] = n_insertions
        metrics["n_deletions"] = n_deletions
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"deploy": metrics})
        self.settings.save()
        self.logger.info("Completed deployment of Seal5 LLVM")

    def export(self, dest: Path, verbose: bool = False, temp: bool = False):
        """Export Seal5 artifacts."""
        del verbose  # unused
        self.logger.info("Exporting Seal5 artifacts")
        start = time.time()
        metrics = {}
        assert dest is not None
        if isinstance(dest, str):
            dest = Path(dest)
        suffix = dest.suffix
        if suffix != ".gz":
            raise NotImplementedError("Only .tar.gz export is supported!")
        artifacts = [
            self.settings.inputs_dir,
            self.settings.gen_dir,
            self.settings.patches_dir,
            self.settings.inputs_dir,
            self.settings.tests_dir,
            self.settings.models_dir,
            self.settings.logs_dir,
            self.settings.settings_file,
        ]
        if temp:
            artifacts.append(self.settings.temp_dir)
        with tarfile.open(dest, mode="w:gz") as archive:
            for artifact in artifacts:
                name = str(artifact)
                assert str(self.meta_dir) in name
                name = name.replace(f"{self.meta_dir}/", "")
                if artifact.is_file():
                    archive.add(artifact, arcname=name)
                elif artifact.is_dir():
                    archive.add(artifact, arcname=name, recursive=True)

        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"export": metrics})
        self.settings.save()
        self.logger.info("Completed export of Seal5 artifacts")

    def reset(self, settings: bool = True, verbose: bool = False, interactive: bool = False):
        """Reset Seal5 flow."""
        del verbose  # unused
        self.logger.info("Cleaning Seal5 state")
        start = time.time()
        metrics = {}
        if interactive:
            raise NotImplementedError
        if settings:
            self.settings.reset()
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"reset": metrics})
        if self.meta_dir.is_dir():
            self.settings.save()
        self.logger.info("Completed clean of Seal5 settings")

    def clean(
        self,
        temp: bool = False,
        patches: bool = False,
        models: bool = False,
        inputs: bool = False,
        logs: bool = False,
        install: bool = False,
        build: bool = False,
        deps: bool = False,
        verbose: bool = False,
        interactive: bool = False,
    ):
        """Cleanup Seal5 flow."""
        del verbose  # unused
        self.logger.info("Cleaning Seal5 directories")
        start = time.time()
        metrics = {}
        to_clean = []
        if temp:
            to_clean.append(self.settings.temp_dir)
        if patches:
            to_clean.append(self.settings.patches_dir)
        if models:
            to_clean.append(self.settings.models_dir)
        if inputs:
            to_clean.append(self.settings.inputs_dir)
        if logs:
            to_clean.append(self.settings.logs_dir)
        if install:
            to_clean.append(self.settings.install_dir)
        if build:
            to_clean.append(self.settings.build_dir)
        if deps:
            to_clean.append(self.settings.deps_dir)
        # TODO: cleanup settings.test.paths or self.settings.tests_dir
        # if gen:
        #     to_clean.append(self.settings.gen_dir)
        for path in to_clean:
            utils.clean_path(path, interactive=interactive)
        # self.reset(verbose=verbose, interactive=interactive)
        end = time.time()
        diff = end - start
        metrics["start"] = start
        metrics["end"] = end
        metrics["time_s"] = diff
        self.settings.metrics.append({"clean": metrics})
        if self.meta_dir.is_dir():
            self.settings.save()
        self.logger.info("Completed clean of Seal5 directories")

    def close_servers(self):
        stop_logging_server()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Detect available registers for Seal5."""

import sys
import argparse
import logging
import pathlib

import seal5.model as seal5_model
from seal5.model_utils import load_model, dump_model

from seal5.logging import Logger

logger = Logger("transform.detect_registers")


def detect_registers(set_def: seal5_model.Seal5InstructionSet):
    to_skip = ["FENCE", "RES", "PRIV", "DPC"]  # TODO: eliminate unused memories during optimization?
    for mem_name, mem in set_def.memories.items():
        if mem_name in to_skip:
            continue
        if mem.is_main_mem or mem.is_pc:
            continue
        assert mem_name not in set_def.registers, "Registers already added to set."
        name = mem_name
        # print("name", name)
        size = mem.data_range.length
        # print("size", size)
        width = mem.size
        # print("width", width)
        # TODO: implement memory dtype in m2isar
        # assert mem.data_type in [arch.DataType.S, arch.DataType.U],
        #  f"Unsupported dtype {mem.data_type} for memory {mem_name}"
        # signed = mem.data_type == arch.DataType.S
        signed = False
        is_const = False
        reg_class = seal5_model.Seal5RegisterClass.CUSTOM
        if name == "X":
            reg_class = seal5_model.Seal5RegisterClass.GPR
        elif name == "F":
            reg_class = seal5_model.Seal5RegisterClass.FPR
        elif name == "CSR":
            reg_class = seal5_model.Seal5RegisterClass.CSR
        if size is None:
            raise NotImplementedError(f"Register with unknown size: {name}")
        if size > 1:  # TODO: check this!
            names = [f"{name}{i}" for i in range(mem.data_range.lower, mem.data_range.upper + 1)]
            # print("names", names)
            group = seal5_model.Seal5RegisterGroup(names, 1, width, signed, is_const, reg_class)
            # print("group", group)
            set_def.register_groups[name] = group
            for reg in group.registers:
                set_def.registers[reg.name] = reg
        else:
            reg = seal5_model.Seal5Register(name, 1, width, signed, is_const, reg_class)
            # print("reg", reg)
            set_def.registers[name] = reg


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    to_skip = ["Zicsr"]  # TODO: use yaml!
    for set_name, set_def in model_obj.sets.items():
        if set_name in to_skip:
            continue
        logger.info("processing set %s", set_name)
        detect_registers(set_def)

    dump_model(model_obj, out_path, compat=args.compat)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Test results report writer for Seal5."""

import argparse
import logging
import pathlib
from collections import defaultdict

import pandas as pd

from seal5.settings import Seal5Settings
from seal5.model import Seal5InstrAttribute
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.test_results_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", nargs="+", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md"], default="auto")
    parser.add_argument("--yaml", type=str, default=None)
    parser.add_argument("--compact", action="store_true")
    parser.add_argument("--markdown-icons", action="store_true")
    parser.add_argument("--coverage", default=None)
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # if args.output is None:
    #     assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
    #     raise NotImplementedError

    #     # out_path = top_level.parent / (top_level.stem + ".core_desc")
    # else:
    assert args.output is not None
    out_path = pathlib.Path(args.output)

    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    def map_tests(metrics):
        test_metrics = None
        for stages_metrics in metrics:
            assert len(stages_metrics) == 1
            if "test" in stages_metrics.keys():
                test_metrics = stages_metrics["test"]
                break
        if test_metrics is None:
            logger.warning("Could not find test metrics. Make sure to run TEST stage!")
            return {}, []
        ret = defaultdict(list)
        ret2 = []
        failing = test_metrics.get("failing", None)
        assert failing is not None, "Missing settings.metrics.test.failing"

        def parse_test_filename(fname):
            fname = pathlib.Path(fname)
            suffix = fname.suffix
            assert len(suffix) > 1
            test_fmt = suffix[1:]
            stem = fname.stem
            if stem.count(".") != 1:
                return None
            instr_name, rest = stem.split(".", 1)
            if not rest.startswith("test"):
                return None
            test_type = rest[5:]
            if len(test_type) == 0:
                test_type = None
            return instr_name, test_type, test_fmt

        for test_file in failing:
            parsed = parse_test_filename(test_file)
            if parsed is None:
                logger.warning("Could no infer instruction from test file name")
                new = (test_file, "FAIL")
                ret2.append(new)
                continue
            instr_name, test_type, test_fmt = parsed
            new = (test_type, test_file, test_fmt, "FAIL")
            ret[instr_name].append(new)
        passed = test_metrics.get("passed", None)
        assert passed is not None, "Missing settings.metrics.test.passed"
        for test_file in passed:
            parsed = parse_test_filename(test_file)
            if parsed is None:
                logger.warning("Could no infer instruction from test file name")
                new = (test_file, "PASS")
                ret2.append(new)
                continue
            instr_name, test_type, test_fmt = parsed
            new = (test_type, test_file, test_fmt, "PASS")
            ret[instr_name].append(new)
        return ret, ret2

    metrics = settings.metrics
    instr_tests, other_tests = map_tests(metrics)
    used_keys = set()

    def filter_tests_by_instr(tests, instr_def):
        instr_name = instr_def.name
        mnemonic = instr_def.mnemonic
        assert isinstance(tests, dict)
        candidates = set()
        candidates.add(instr_name)
        candidates.add(instr_name.lower())
        candidates.add(mnemonic)
        candidates.add(mnemonic.lower())
        instr_name_alt = instr_name.replace("seal5.", "").replace("SEAL5.", "")
        instr_name_alt = instr_name_alt.replace("seal5_", "").replace("SEAL5_", "")
        instr_name_alt = instr_name_alt.replace(".", "-")
        candidates.add(instr_name_alt)
        candidates.add(instr_name_alt.lower())
        mnemonic_alt = mnemonic.replace("seal5.", "").replace("SEAL5.", "")
        mnemonic_alt = mnemonic_alt.replace("seal5_", "").replace("SEAL5_", "")
        mnemonic_alt = mnemonic_alt.replace(".", "-")
        candidates.add(mnemonic_alt)
        candidates.add(mnemonic_alt.lower())
        instr_name_alt = instr_name_alt.replace("_", "-")
        candidates.add(instr_name_alt)
        candidates.add(instr_name_alt.lower())
        mnemonic_alt = mnemonic_alt.replace("_", "-")
        candidates.add(mnemonic_alt)
        candidates.add(mnemonic_alt.lower())

        ret = []
        used = set()
        for candidate_name in candidates:
            ret_ = tests.get(candidate_name, None)
            if ret_ is not None:
                assert isinstance(ret_, list)
                ret.extend(ret_)
                used.add(candidate_name)
        return ret, used

    def filter_tests_by_set(tests, set_def, settings, model_name):
        set_name = set_def.name
        assert isinstance(tests, dict)
        candidates = set()
        candidates.add(set_name)
        candidates.add(set_name.lower())
        set_name_alt = set_name.replace("_", "")
        candidates.add(set_name_alt)
        candidates.add(set_name_alt.lower())
        if settings:
            model_settings = settings.models[model_name]
            extension_settings = model_settings.extensions.get(set_name)
            if extension_settings:
                arch = extension_settings.get_arch(set_name)
                feature = extension_settings.get_feature(set_name)
                candidates.add(arch)
                candidates.add(feature)
                candidates.add(arch.lower())
                candidates.add(feature.lower())
                arch_alt = arch.replace("_", "")
                feature_alt = feature.replace("_", "")
                candidates.add(arch_alt)
                candidates.add(feature_alt)
                candidates.add(arch_alt.lower())
                candidates.add(feature_alt.lower())

        ret = []
        used = set()
        for candidate_name in candidates:
            ret_ = tests.get(candidate_name, None)
            if ret_ is not None:
                assert isinstance(ret_, list)
                ret.extend(ret_)
                used.add(candidate_name)
        return ret, used

    results_data = []
    if args.coverage:
        coverage_data = []
    # resolve model paths
    top_levels = args.top_level
    for top_level in top_levels:
        top_level = pathlib.Path(top_level)

        model_obj = load_model(top_level, compat=args.compat)

        for set_name, set_def in model_obj.sets.items():
            if len(set_def.instructions) == 0:
                continue
            xlen = set_def.xlen
            model_name = top_level.stem

            for instr_def in set_def.instructions.values():
                instr_name = instr_def.name

                data = {
                    "model": model_name,
                    "set": set_name,
                    "xlen": xlen,
                    "instr": instr_name,
                }

                found_tests, used = filter_tests_by_instr(instr_tests, instr_def)
                if args.coverage:

                    def get_instr_cov(instr_def, tests, settings):
                        skip_pattern_gen = Seal5InstrAttribute.SKIP_PATTERN_GEN in instr_def.attributes
                        has_legalizer_settings = False  # TODO: get from settings

                        def lookup_intrin(instr_name, settings):
                            if settings.intrinsics is None:
                                return None
                            intrinsics = settings.intrinsics.intrinsics
                            if not intrinsics:
                                return None
                            found = None
                            for intrin in intrinsics:
                                if intrin.instr_name == instr_name:
                                    found = intrin
                                    break
                            return found

                        intrin = lookup_intrin(instr_def.name, settings)
                        has_intrin = intrin is not None
                        # (test_kind, test_fmt): optional
                        temp = {
                            ("mc", "s"): True,
                            ("mc-invalid", "s"): False,
                            ("inline-asm", "c"): False,
                            ("cg", "ll"): not skip_pattern_gen,
                            ("cg", "c"): not skip_pattern_gen,
                            ("cg", "mir"): not skip_pattern_gen,
                            ("legalizer", "mir"): has_legalizer_settings,
                            ("intrin", "ll"): has_intrin,
                            ("builtin", "c"): has_intrin,
                        }
                        # (test_kind, test_fmt, optional): (extra, exists)
                        temp2 = {key: (val, False, False) for key, val in temp.items()}
                        for test_kind, test_file, test_fmt, test_result in found_tests:
                            required = temp.get((test_kind, test_fmt), None)
                            if required is not None:
                                temp2[(test_kind, test_fmt)] = (required, False, True)
                            else:
                                temp2[(test_kind, test_fmt)] = (False, True, True)
                        # (test_kind, test_fmt, optional, extra, exists)
                        ret = [(*key, *val) for key, val in temp2.items()]
                        return ret

                    instr_cov = get_instr_cov(instr_def, found_tests, settings)
                    for test_kind, test_fmt, required, extra, exists in instr_cov:

                        def helper(required, extra, exists):
                            if exists:
                                return "good"
                            if not required:
                                return "ok"
                            return "bad"

                        data_cov = {
                            **data,
                            "test_kind": test_kind,
                            "test_fmt": test_fmt,
                            "required": required,
                            "extra": extra,
                            "exists": exists,
                            "coverage": helper(required, extra, exists),
                        }
                        coverage_data.append(data_cov)
                used_keys.update(used)
                for test_kind, test_file, test_fmt, test_result in found_tests:
                    data_ = {
                        **data,
                        "test_kind": test_kind,
                        "test_file": test_file,
                        "test_fmt": test_fmt,
                        "result": test_result,
                    }
                    results_data.append(data_)
            found_tests, used = filter_tests_by_set(instr_tests, set_def, settings, model_name)
            used_keys.update(used)
            data["instr"] = None
            for test_kind, test_file, test_fmt, test_result in found_tests:
                data_ = {
                    **data,
                    "test_kind": test_kind,
                    "test_file": test_file,
                    "test_fmt": test_fmt,
                    "result": test_result,
                }
                results_data.append(data_)

    def get_unused_tests(tests, used_keys):
        ret = sum([val for key, val in tests.items() if key not in used_keys], [])
        return ret

    unused_tests = get_unused_tests(instr_tests, used_keys)

    data = {
        "model": None,
        "set": None,
        "xlen": None,
        "instr": None,
    }

    for test_kind, test_file, test_fmt, test_result in unused_tests:
        data_ = {
            **data,
            "test_kind": test_kind,
            "test_file": test_file,
            "test_fmt": test_fmt,
            "result": test_result,
        }
        results_data.append(data_)

    for test_file, test_result in other_tests:
        test_fmt = pathlib.Path(test_file).suffix[1:]
        test_kind = None
        data_ = {
            **data,
            "test_kind": test_kind,
            "test_file": test_file,
            "test_fmt": test_fmt,
            "result": test_result,
        }
        results_data.append(data_)

    results_df = pd.DataFrame(results_data)

    if args.compact and len(results_df) > 0:
        new = []
        for group, group_df in results_df.groupby(["model", "set", "xlen", "instr"], dropna=False):
            n_tests = len(group_df)
            counts = group_df["result"].value_counts()
            n_pass = counts.get("PASS", 0)
            n_fail = counts.get("FAIL", 0)

            def helper(n_tests, n_pass, n_fail):
                if n_fail:
                    return "bad"
                if n_pass:
                    return "good"
                return "unknown"

            state = helper(n_tests, n_pass, n_fail)
            new_ = {
                "model": group[0],
                "set": group[1],
                "xlen": group[2],
                "instr": group[3],
                "n_pass": n_pass,
                "n_fail": n_fail,
                "n_tests": n_tests,
                "state": state,
            }
            new.append(new_)
        results_df = pd.DataFrame(new)

    if args.coverage:
        # TODO: also add set test coverage
        coverage_df = pd.DataFrame(coverage_data)
        if args.compact and len(coverage_df) > 0:
            new = []
            for group, group_df in coverage_df.groupby(["model", "set", "xlen", "instr"], dropna=False):
                n_exists = group_df["exists"].value_counts().get(True, 0)
                n_required = group_df["required"].value_counts().get(True, 0)
                n_optional = group_df[["required", "extra"]].value_counts().get((False, False), 0)
                n_extra = group_df["extra"].value_counts().get(True, 0)
                n_required_exists = group_df[["required", "exists"]].value_counts().get((True, True), 0)
                n_optional_exists = group_df[["required", "exists"]].value_counts().get((False, True), 0)

                def helper2(n_exists, n_required, n_required_exists, n_optional, n_optional_exists):
                    if n_exists == 0:
                        return "bad"
                    if n_required_exists < n_required:
                        return "bad"
                    assert n_required_exists == n_required
                    if n_optional_exists == n_optional:
                        return "good"
                    return "ok"

                cov = helper2(n_exists, n_required, n_required_exists, n_optional, n_optional_exists)
                new_ = {
                    "model": group[0],
                    "set": group[1],
                    "xlen": group[2],
                    "instr": group[3],
                    "n_exists": n_exists,
                    "n_required": n_required,
                    "n_optional": n_optional,
                    "n_extra": n_extra,
                    "n_required_exists": n_required_exists,
                    "n_optional_exists": n_optional_exists,
                    "coverage": cov,
                }
                new.append(new_)
            coverage_df = pd.DataFrame(new)

    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        results_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        results_df.to_pickle(out_path)
    elif fmt == "md":
        # results_df.to_markdown(out_path, tablefmt="grid", index=False)
        if args.markdown_icons and len(results_df) > 0:

            def helper(x):
                x = x.replace("PASS", "PASS :heavy_check_mark:")
                x = x.replace("FAIL", "FAIL :x:")
                x = x.replace("good", "good :green_circle:")
                x = x.replace("ok", "ok :orange_circle:")
                x = x.replace("unknown", "unknown :yellow_circle:")
                x = x.replace("bad", "bad :red_circle:")
                return x

            if args.compact:
                results_df["state"] = results_df["state"].apply(helper)
            else:
                results_df["result"] = results_df["result"].apply(helper)
        results_df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")

    if args.coverage:
        # TODO: share code
        cov_path = pathlib.Path(args.coverage)
        fmt = args.fmt
        if fmt == "auto":
            fmt = cov_path.suffix
            assert len(fmt) > 1
            fmt = fmt[1:].lower()

        if fmt == "csv":
            coverage_df.to_csv(cov_path, index=False)
        elif fmt == "pkl":
            coverage_df.to_pickle(cov_path)
        elif fmt == "md":
            # coverage_df.to_markdown(cov_path, tablefmt="grid", index=False)
            if args.markdown_icons and len(coverage_df) > 0:

                def helper(x):
                    x = x.replace("good", "good :green_circle:")
                    x = x.replace("ok", "ok :orange_circle:")
                    x = x.replace("unknown", "unknown :yellow_circle:")
                    x = x.replace("bad", "bad :red_circle:")
                    return x

                if args.compact:
                    coverage_df["coverage"] = coverage_df["coverage"].apply(helper)
                else:
                    coverage_df["coverage"] = coverage_df["coverage"].apply(helper)
            coverage_df.to_markdown(cov_path, index=False)
        else:
            raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Demo script for the Seal5 Flow."""
import os

from pathlib import Path

from seal5.wrapper import run_seal5_flow


DEMO_NAME = os.environ.get("DEMO_NAME", "s4e")
EXAMPLES_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
DEMO_CDSL_DIR = Path(os.environ.get("DEMO_CDSL_DIR", EXAMPLES_DIR / DEMO_NAME / "cdsl"))
DEMO_TESTS_DIR = Path(os.environ.get("DEMO_TESTS_DIR", EXAMPLES_DIR / DEMO_NAME / "tests"))
DEMO_CFG_DIR = Path(os.environ.get("DEMO_CFG_DIR", EXAMPLES_DIR / DEMO_NAME / "cfg"))
COMMON_CFG_DIR = Path(os.environ.get("COMMON_CFG_DIR", EXAMPLES_DIR / "common" / "cfg"))
DEST_DIR = os.environ.get("DEST_DIR", "/tmp")
DEST = os.environ.get("DEST", DEST_DIR + "/seal5_llvm_" + DEMO_NAME).rstrip("/")

FILES = [
    # CoreDSL inputs
    DEMO_CDSL_DIR / "s4e-mac.core_desc",
    # Test inputs
    DEMO_TESTS_DIR / "*.s",
    DEMO_TESTS_DIR / "*.c",
    # DEMO_TESTS_DIR / "*.ll",
    # YAML inputs
    DEMO_CFG_DIR / "s4e-mac.yml",
    COMMON_CFG_DIR / "llvm.yml",
    COMMON_CFG_DIR / "filter.yml",
    COMMON_CFG_DIR / "patches.yml",
    COMMON_CFG_DIR / "riscv.yml",
    COMMON_CFG_DIR / "tests.yml",
    COMMON_CFG_DIR / "passes.yml",
    COMMON_CFG_DIR / "git.yml",
]
run_seal5_flow(FILES, name=DEMO_NAME, dest=DEST)



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Patch utils for seal5."""
import os
import argparse
from pathlib import Path
from typing import Optional

from email.utils import formatdate
import yaml

from seal5 import utils
from seal5.logging import Logger

logger = Logger("tools")


def generate_patch(index_file, llvm_dir=None, out_file=None, author=None, mail=None, msg=None, append=None):
    """Generate patch contents based on index file."""
    del append  # unused
    # base_dir = os.path.dirname(index_file)
    if msg:
        if not isinstance(msg, str):
            assert isinstance(msg, list)
            assert len(msg) == 1
            msg = msg[0]
    else:
        msg = "Placeholder"

    with open(index_file, "r", encoding="utf-8") as file:
        index = yaml.safe_load(file)

    global_artifacts = index["artifacts"]

    def generate_patch_set(fragments):
        ps = ""
        # Build minimal patch header for 'git am'
        user_name = author
        user_email = mail
        ps += f"From: {user_name} <{user_email}>\n"
        ps += f"Date: {formatdate()}\n"
        ps += f"Subject: {msg}\n\n\n"
        for fragment in fragments:
            ps += fragment
        return ps

    def find_site(path, key):
        # print("find_site", path, key)
        # start_mark = key + ' - INSERTION_START'
        # enc_mark = key + ' - INSERTION_END'
        if key:
            start_mark_ = path.split("/")[-1] + " - " + key + " - INSERTION_START"
            end_mark_ = path.split("/")[-1] + " - " + key + " - INSERTION_END"
        else:
            start_mark_ = path.split("/")[-1] + " - INSERTION_START"
            end_mark_ = path.split("/")[-1] + " - INSERTION_END"
        # print("start", start_mark_)
        # print("end", end_mark_)
        site_line = -1
        site_len = 0
        start_mark = None
        fullpath = path if llvm_dir is None else os.path.join(llvm_dir, path)
        with open(fullpath, "r", encoding="utf-8") as src:
            lines = src.readlines()
            for line in lines:
                if line.find(start_mark_) != -1:
                    start_mark = line
                    site_line = lines.index(line)
                elif line.find(end_mark_) != -1:
                    end_mark = line.rstrip("\n")
                    site_len = lines.index(line) - site_line + 1
                    break
                elif site_line >= 0:
                    # Accumulate all lines between the start mark and the end
                    # as part of the mark, so the new lines get injected just
                    # before the end mark
                    start_mark += line
        if start_mark is None:
            # fallback
            logger.warning("Marker not found: {path}, {key}. Retrying without key...")
            if key:
                return find_site(path, None)
            assert False, f"Marker not found: {path}, {key}"
        return site_line, site_len, start_mark, end_mark

    def generate_patch_fragment(artifact):
        dest_path = artifact.get("dest_path", None)
        src_path = artifact.get("src_path", None)
        key = artifact.get("key", None)
        start = artifact.get("start", None)
        end = artifact.get("end", None)
        line = artifact.get("line", None)
        content_ = artifact.get("content", None)
        assert dest_path is not None
        is_file = False
        is_dir = False
        is_patch = False
        # TODO: do not depend on dest_path
        if key:  # NamedPatch
            is_patch = True
        elif start is not None and end is not None:
            raise NotImplementedError
        elif line is not None:
            raise NotImplementedError
        else:  # File/Dir
            if content_ is None:
                if Path(src_path).is_file():
                    is_file = True
                elif Path(src_path).is_dir():
                    is_dir = True
                else:
                    raise RuntimeError(f"File not found: {src_path}")
            else:
                is_file = True
        if content_:
            assert isinstance(content_, str)
        else:
            if is_patch or is_file:
                with open(src_path, "r", encoding="utf-8") as f:
                    content_ = f.read()
            elif is_dir:
                files = Path(src_path).rglob("*")
                base = dest_path
                # print("base", base)
                # print("files", files)
                fragments = []
                for file in files:
                    if file.is_dir():
                        continue
                    # print("file", file)
                    dest_path_ = str(Path(base) / (str(file).replace(f"{src_path}/", "")))
                    # print("dest_path_", dest_path_)
                    src_path_ = file
                    file_artifact = {
                        "dest_path": str(dest_path_),
                        "src_path": str(src_path_),
                    }
                    fragment = generate_patch_fragment(file_artifact)
                    fragments.append(fragment)
                # print("fragments", fragments)
                # input("AAA")
                return "".join(fragments)
                # raise NotImplementedError
            else:
                assert False
        content = "+" + content_.replace("\n", "\n+")
        if is_patch:
            # Updating existing file
            orig_file = "a/" + dest_path
            new_file = "b/" + dest_path
            site_line, site_len, start_mark, end_mark = find_site(dest_path, key)
            new_start = site_line + 1
            new_len = content_.count("\n") + 1 + site_len
            # ensure all existing lines in the match prefixed by a space
            # if site_len > 2 and start_mark[-1] == "\n":
            # if False:
            #     start_mark = start_mark.rstrip("\n").replace("\n", "\n ")
            # else:
            start_mark = start_mark.rstrip("\n").replace("\n", "\n ")
            content = f" {start_mark}\n{content}\n {end_mark}"
        else:
            assert is_file
            # Adding new file
            orig_file = "/dev/null"
            new_file = "b/" + dest_path
            new_start = 1
            site_len = 0
            site_line = -1
            new_len = content_.count("\n") + 1
        return f"""--- {orig_file}
+++ {new_file}
@@ -{site_line + 1},{site_len} +{new_start},{new_len} @@
{content}
"""

    fragments = []

    def process_artifacts(artifacts):
        ret = []
        for artifact in artifacts:
            fragment = generate_patch_fragment(artifact)
            ret.append(fragment)
        return ret

    fragments.extend(process_artifacts(global_artifacts))

    for extn in index["extensions"]:
        fragments.extend(process_artifacts(extn["artifacts"]))

    patch_set = generate_patch_set(fragments)

    if out_file:
        with open(out_file, "w", encoding="utf-8") as f:
            f.write(patch_set)
    else:
        print(patch_set)


def process_arguments():
    """Progress cmdline arguments."""
    parser = argparse.ArgumentParser(description="Generate patch file for adding extension to LLVM")
    parser.add_argument("index_file", help="list of code fragments to inject")
    parser.add_argument("--llvm-dir", type=str, default=None, help="path to llvm repo (default: cwd)")
    parser.add_argument("--out-file", type=str, default=None, help="write patchset to file (default: print to stdout)")
    parser.add_argument("--author", type=str, default="bob", help="author name in patchset")
    parser.add_argument("--mail", type=str, default="bob@bob.com", help="author mail in patchset")
    parser.add_argument(
        "--msg", type=str, nargs="+", default="[PATCH] Instructions injected", help="custom text for commit message"
    )
    parser.add_argument(
        "--append",
        "-a",
        action="store_true",
        help="whether existing patches should be appended instead of beeing overwritten",
    )
    args = parser.parse_args()
    return args


def analyze_diff(repo, base: str, cur: Optional[str] = None):
    args = ["git", "diff", "--shortstat"]
    args += [base]
    if cur is not None:
        args += [cur]
    out = utils.exec_getout(
        *args,
        cwd=repo.working_tree_dir,
        print_func=lambda *args, **kwargs: None,
        live=False,
    )
    n_files_changed = 0
    n_insertions = 0
    n_deletions = 0
    for x in out.split(","):
        x = x.strip()
        if len(x) == 0:
            continue
        val, key = x.split(" ", 1)
        val = int(val)
        if "files changed" in key or "file changed" in key:
            n_files_changed = val
        elif "insertions" in key:
            n_insertions = val
        elif "deletions" in key:
            n_deletions = val
    return n_files_changed, n_insertions, n_deletions


def main():
    """Main entry point."""
    args = process_arguments()
    generate_patch(
        args.index_file,
        llvm_dir=args.llvm_dir,
        out_file=args.out_file,
        author=args.authot,
        mail=args.mail,
        msg=args.msg,
        append=args.append,
    )


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Seal5 TestGen Utils."""


def collect_generated_test_files(index_file):
    import yaml

    with open(index_file, "r", encoding="utf-8") as file:
        index = yaml.safe_load(file)

    global_artifacts = index["artifacts"]
    ext_artifacts = sum([ext["artifacts"] for ext in index["extensions"]], [])
    all_artifacts = global_artifacts + ext_artifacts
    ret = []
    for artifact in all_artifacts:
        if artifact.get("is_test", False):
            dest_path = artifact["dest_path"]
            if dest_path.startswith("llvm/test/"):
                dest_path = dest_path.replace("llvm/test/", "")
            elif dest_path.startswith("./llvm/test/"):
                dest_path = dest_path.replace("./llvm/test/", "")
            ret.append(dest_path)
    return ret



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav

# pylint: disable=unused-argument

# Limitations
# - Keep the condition stack as small as possible (move raise to highest level of behavior)
# - If block should only have a call to raise
# - Standalone if statements (without else/else if) are preferred
# - Raise calls are replaced by empty blocks, to be removed in the operation visitor
# - ~~Run optimizer afterwards to eliminate if statement.~~


def operation(self: behav.Operation, context):
    # print("operation", self)
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            # for t in temp:
            # print("t", t, type(t), dir(t))
            statements.extend(temp)
        else:
            # print("temp", temp, type(temp), dir(temp))
            statements.append(temp)
    # input("eeeee")

    self.statements = statements
    return self


def block(self: behav.Block, context):
    # print("block", self)

    stmts = []

    for stmt in self.statements:
        stmt = stmt.generate(context)
        if isinstance(stmt, behav.Conditional):
            if len(stmt.conds) == 1:
                if isinstance(stmt.stmts[0], behav.Block):
                    if len(stmt.stmts[0].statements) == 0:
                        continue
        stmts.append(stmt)

    self.statements = stmts

    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    return self


def conditional(self: behav.Conditional, context):
    # print("conditional")
    self.conds = [x.generate(context) for x in self.conds]
    # self.stmts = [x.generate(context) for x in self.stmts]
    temp = []
    for i, stmt in enumerate(self.stmts):
        if i == 0:  # if
            cond = self.conds[i]
            temp.append(behav.UnaryOperation(behav.Operator("!"), cond))
        elif i < len(self.conds):  # elif
            cond = temp[0]
            for c in temp[1:]:
                cond = behav.BinaryOperation(cond, "&&", c)
            cond = behav.BinaryOperation(cond, "&&", self.conds[i])
        else:  # else
            assert i >= len(self.conds)
            assert len(temp) >= 1
            cond = temp[0]
            for c in temp[1:]:
                cond = behav.BinaryOperation(cond, "&&", c)
        context.cond_stack.append(cond)
        # print("before", context.cond_stack)
        stmt = stmt.generate(context)
        if context.found_raise:
            # print("stmt", stmt, dir(stmt))
            if isinstance(stmt, behav.Block):
                # print("block.statements", stmt.statements)
                assert len(stmt.statements) == 1, "Raise block may only have a single operation"
                assert isinstance(stmt.statements[0], behav.ProcedureCall), "Nesting raises not allowed"
                assert isinstance(stmt.statements[0].ref_or_name, arch.Function), "Expected function"
                assert stmt.statements[0].ref_or_name.name == "raise", "Expected raise operation"
                stmt = behav.Block([])  # Replace with empty block
                context.found_raise = False
            else:
                assert isinstance(stmt, behav.ProcedureCall), "Nesting raises not allowed"
                assert isinstance(stmt.ref_or_name, arch.Function), "Expected function"
                assert stmt.ref_or_name.name == "raise", "Expected raise operation"
                stmt = behav.Block([])  # Replace with empty block
                context.found_raise = False
            # input("aaaa")
        self.stmts[i] = stmt
        # print("after", context.cond_stack)
        context.cond_stack.pop()
    # print("self.stmts", self.stmts)

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    return self


def indexed_reference(self: behav.IndexedReference, context):
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    # print("group", group)
    self.expr = self.expr.generate(context)

    return self


def procedure_call(self: behav.ProcedureCall, context):
    # print("procedure_call")

    fn_args = [arg.generate(context) for arg in self.args]

    # extract function object reference
    ref = self.ref_or_name if isinstance(self.ref_or_name, arch.Function) else None
    name = ref.name if isinstance(self.ref_or_name, arch.Function) else self.ref_or_name
    # print("name", name)
    # print("args", fn_args)
    if name == "raise":
        args = [arg.value for arg in fn_args]
        if len(context.cond_stack) == 0:
            cond = None
        else:
            cond = context.cond_stack[-1]
            # for c in context.cond_stack[1:]:
            cond = behav.UnaryOperation(behav.Operator("!"), behav.Group(cond))
            # TODO: transform !(imm % 4) -> !(imm % 4 != 0) -> imm % 4 == 0
        context.raises.append((cond, args))
        context.found_raise = True
        # input("qqqq")
    return self



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for initializing Seal5 environment."""

from seal5.flow import Seal5Flow
from seal5.logging import Logger


logger = Logger("cli")


def add_init_options(parser):
    """Setup parser for init argument group."""
    init_parser = parser.add_argument_group("init options")
    init_parser.add_argument(
        "--non-interactive",
        action="store_true",
        help="Do not ask questions interactively",
    )
    init_parser.add_argument(
        "--clone",
        "-c",
        action="store_true",
        help="Clone LLVM repository",
    )
    init_parser.add_argument(
        "--clone-url",
        default="https://github.com/llvm/llvm-project.git",
        help="Corresponding LLVM repository URL",
    )
    init_parser.add_argument(
        "--clone-ref",
        default="llvmorg-19.1.0",
        help="Corresponding LLVM repository commit/tag",
    )
    init_parser.add_argument(
        "--clone-depth",
        default=None,
        type=int,
        help="LLVM clone depth (use 1 for shallow clone)",
    )
    init_parser.add_argument(
        "--progress",
        action="store_true",
        help="Show progress bar during LLVM clone",
    )
    init_parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Allow overwriting an existing seal5 directory",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the init subcommand."""
    parser = subparsers.add_parser("init", description="Initialize Seal5.")
    parser.set_defaults(func=handle)
    add_init_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the init subcommand"""
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.initialize(
        interactive=not args.non_interactive,
        clone=args.clone,
        clone_url=args.clone_url,
        clone_ref=args.clone_ref,
        clone_depth=args.clone_depth if isinstance(args.clone_depth, int) and args.clone_depth > 0 else None,
        progress=args.progress,
        force=args.force,
        verbose=args.verbose,
    )



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav

# pylint: disable=unused-argument

# Limitations
# - Keep the condition stack as small as possible (move raise to highest level of behavior)
# - If block should only have a call to raise
# - Standalone if statements (without else/else if) are preferred
# - Raise calls are replaced by empty blocks, to be removed in the operation visitor
# - ~~Run optimizer afterwards to eliminate if statement.~~


def operation(self: behav.Operation, context):
    # print("operation", self)
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            for _t in temp:
                pass
                # print("t", t, type(t), dir(t))
            statements.extend(temp)
        else:
            # print("temp", temp, type(temp), dir(temp))
            statements.append(temp)
    # input("eeeee")

    self.statements = statements
    return self


def block(self: behav.Block, context):
    # print("block", self)

    stmts = []

    for stmt in self.statements:
        stmt = stmt.generate(context)
        if isinstance(stmt, behav.Conditional):
            if len(stmt.conds) == 1:
                if isinstance(stmt.stmts[0], behav.Block):
                    if len(stmt.stmts[0].statements) == 0:
                        continue
        stmts.append(stmt)

    self.statements = stmts

    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    # print("assignment")
    context.is_write = True
    self.target = self.target.generate(context)
    context.is_write = False
    context.is_read = True
    self.expr = self.expr.generate(context)
    context.is_read = False

    return self


def conditional(self: behav.Conditional, context):
    # print("conditional")
    for i, cond in enumerate(self.conds):
        context.is_read = True
        self.conds[i] = cond.generate(context)
        context.is_read = False
    for op in self.stmts:
        # for stmt in flatten(op):
        #    stmt.generate(context)
        op.generate(context)

    return self


def loop(self: behav.Loop, context):
    context.is_read = True
    self.cond = self.cond.generate(context)
    context.is_read = False
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    context.is_read = True
    self.cond = self.cond.generate(context)
    context.is_read = False
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    return self


def indexed_reference(self: behav.IndexedReference, context):
    # print("indexed_reference")
    assert isinstance(self.reference, arch.Memory)
    # name = self.reference.name
    if arch.MemoryAttribute.IS_MAIN_MEM in self.reference.attributes:
        if context.is_write:
            context.may_store = True
        elif context.is_read:
            context.may_load = True
    else:
        pass  # TODO

    self.index = self.index.generate(context)
    return self


def type_conv(self: behav.TypeConv, context):
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    # print("group", group)
    self.expr = self.expr.generate(context)

    return self


def procedure_call(self: behav.ProcedureCall, context):
    # print("procedure_call")

    self.fn_args = [arg.generate(context) for arg in self.args]

    return self



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

import itertools
from typing import Union

from m2isar import M2DuplicateError, M2NameError, M2TypeError, M2ValueError, flatten
from m2isar.metamodel import arch, behav, intrinsics
from .parser_gen import CoreDSL2Parser, CoreDSL2Visitor
from .utils import RADIX, SHORTHANDS, SIGNEDNESS

from seal5.logging import Logger

logger = Logger("frontends.arch_builder")


class ArchitectureModelBuilder(CoreDSL2Visitor):
    """ANTLR visitor to build an M2-ISA-R architecture model of a CoreDSL 2 specification."""

    _constants: "dict[str, arch.Constant]"
    _instructions: "dict[str, arch.Instruction]"
    _functions: "dict[str, arch.Function]"
    _always_blocks: "dict[str, arch.AlwaysBlock]"
    _instruction_sets: "dict[str, arch.InstructionSet]"
    _read_types: "dict[str, str]"
    _memories: "dict[str, arch.Memory]"
    _memory_aliases: "dict[str, arch.Memory]"
    _overwritten_instrs: "list[tuple[arch.Instruction, arch.Instruction]]"
    _instr_classes: "set[int]"
    _main_reg_file: Union[arch.Memory, None]

    def __init__(self):
        super().__init__()
        self._constants = {}
        self._instructions = {}
        self._functions = {}
        self._always_blocks = {}
        self._instruction_sets = {}
        self._read_types = {}
        self._memories = {}
        self._memory_aliases = {}

        self._overwritten_instrs = []
        self._instr_classes = set()
        self._main_reg_file = None

    def visitBit_field(self, ctx: CoreDSL2Parser.Bit_fieldContext):
        """Generate a bit field (instruction parameter in encoding)."""

        # generate lower and upper bounds
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)

        # instantiate M2-ISA-R objects
        range = arch.RangeSpec(left.value, right.value)
        return arch.BitField(ctx.name.text, range, arch.DataType.U)

    def visitBit_value(self, ctx: CoreDSL2Parser.Bit_valueContext):
        """Generate a fixed encoding part."""

        val = self.visit(ctx.value)
        return arch.BitVal(val.bit_size, val.value)

    def visitInstruction_set(self, ctx: CoreDSL2Parser.Instruction_setContext):
        """Generate a top-level instruction set object."""
        # print("visitInstruction_set", ctx.name.text)
        # print("len(self._constants) 1", len(self._constants))
        # input("22")

        # keep track of seen instruction set names
        self._read_types[ctx.name.text] = None

        name = ctx.name.text
        extension = []
        if ctx.extension:
            extension = [obj.text for obj in ctx.extension]

        # generate flat list of instruction set contents
        contents = flatten([self.visit(obj) for obj in ctx.sections])
        # print("len(self._constants) 2", len(self._constants))
        # input("33")

        constants = {}
        constants.update(self._constants)
        memories = {}
        memories.update(self._memories)
        functions = {}
        functions.update(self._functions)
        instructions = {}
        # instructions.update(self._instructions)

        # group contents by type
        for item in contents:
            if isinstance(item, arch.Constant):
                constants[item.name] = item
            elif isinstance(item, arch.Memory):
                memories[item.name] = item
            elif isinstance(item, arch.Function):
                functions[item.name] = item
                item.ext_name = name
            elif isinstance(item, arch.Instruction):
                instructions[(item.code, item.mask)] = item
                item.ext_name = name
            elif isinstance(item, arch.AlwaysBlock):
                pass
            else:
                raise M2ValueError("unexpected item encountered")

        # instantiate M2-ISA-R object
        i = arch.InstructionSet(name, extension, constants, memories, functions, instructions)

        if name in self._instruction_sets:
            raise M2DuplicateError(f'instruction set "{name}" already defined')

        # keep track of instruction set object
        self._instruction_sets[name] = i
        return i

    def visitSection_instructions(self, ctx: CoreDSL2Parser.Section_instructionsContext):
        attributes = dict([self.visit(obj) for obj in ctx.attributes])
        instructions: "list[arch.Instruction]" = [self.visit(obj) for obj in ctx.instructions]

        for attr, val in attributes.items():
            for instr in instructions:
                if attr not in instr.attributes:
                    instr.attributes[attr] = val

        return instructions

    def visitCore_def(self, ctx: CoreDSL2Parser.Core_defContext):
        """Generate a top-level CoreDef object."""

        self.visitChildren(ctx)

        name = ctx.name.text

        c = arch.CoreDef(
            name,
            list(self._read_types.keys()),
            None,
            self._constants,
            self._memories,
            self._memory_aliases,
            self._functions,
            self._instructions,
            self._instr_classes,
            intrinsics,
        )

        return c

    def visitSection_arch_state(self, ctx: CoreDSL2Parser.Section_arch_stateContext):
        """Generate "archictectural_state" section of CoreDSL file."""

        decls = [self.visit(obj) for obj in ctx.declarations]
        decls = list(itertools.chain.from_iterable(decls))
        for obj in ctx.expressions:
            self.visit(obj)

        return decls

    def visitAlways_block(self, ctx: CoreDSL2Parser.Always_blockContext):
        """Generate always block"""

        name = ctx.name.text
        attributes = dict([self.visit(obj) for obj in ctx.attributes])

        a = arch.AlwaysBlock(name, attributes, ctx.behavior)

        self._always_blocks[name] = a

        return a

    def visitInstruction(self, ctx: CoreDSL2Parser.InstructionContext):
        """Generate non-behavioral parts of an instruction."""

        # read encoding, attributes and disassembly
        encoding = [self.visit(obj) for obj in ctx.encoding]
        attributes = dict([self.visit(obj) for obj in ctx.attributes])
        assembly = ctx.assembly.text.replace('"', "") if ctx.assembly is not None else None
        mnemonic = ctx.mnemonic.text.replace('"', "") if ctx.mnemonic is not None else None

        i = arch.Instruction(ctx.name.text, attributes, encoding, mnemonic, assembly, ctx.behavior, None)
        self._instr_classes.add(i.size)

        instr_id = (i.code, i.mask)

        # check for duplicate instructions
        if instr_id in self._instructions:
            self._overwritten_instrs.append((self._instructions[instr_id], i))

        # keep track of instruction
        self._instructions[instr_id] = i

        return i

    def visitFunction_definition(self, ctx: CoreDSL2Parser.Function_definitionContext):
        """Generate non-behavioral parts of a function."""

        # decode attributes
        attributes = dict([self.visit(obj) for obj in ctx.attributes])

        if arch.FunctionAttribute.ETISS_TRAP_ENTRY_FN in attributes:
            attributes[arch.FunctionAttribute.ETISS_NEEDS_ARCH] = []

        # decode return type and name
        type_ = self.visit(ctx.type_)
        name = ctx.name.text

        # decode function arguments
        params = []
        if ctx.params:
            params = self.visit(ctx.params)

        if not isinstance(params, list):
            params = [params]

        return_size = None
        data_type = arch.DataType.NONE

        if isinstance(type_, arch.IntegerType):
            return_size = type_._width
            data_type = arch.DataType.S if type_.signed else arch.DataType.U

        f = arch.Function(name, attributes, return_size, data_type, params, ctx.behavior, ctx.extern is not None)

        # error on duplicate function definition
        # TODO: implement overwriting function prototypes?
        f2 = self._functions.get(name, None)

        if f2 is not None:
            if len(f2.operation.statements) > 0:
                raise M2DuplicateError(f'function "{name}" already defined')

            self._functions.pop(name)

        self._functions[name] = f
        return f

    def visitParameter_declaration(self, ctx: CoreDSL2Parser.Parameter_declarationContext):
        """Generate function argument declaration."""

        # type is required, name and array size optional
        type_ = self.visit(ctx.type_)
        name = None
        # size = None
        if ctx.decl:
            if ctx.decl.name:
                name = ctx.decl.name.text
            if ctx.decl.size:
                ctx.decl.size = [self.visit(obj) for obj in ctx.decl.size]

        p = arch.FnParam(name, type_._width, arch.DataType.S if type_.signed else arch.DataType.U)
        return p

    def visitInteger_constant(self, ctx: CoreDSL2Parser.Integer_constantContext):
        """Generate an integer literal."""

        # extract raw text
        text: str = ctx.value.text.lower()

        # extract tick position for verilog-stlye literal
        tick_pos = text.find("'")

        # decode verilog-style literal
        if tick_pos != -1:
            width = int(text[:tick_pos])
            radix = text[tick_pos + 1]
            value = int(text[tick_pos + 2 :], RADIX[radix])

        # decode normal dec, hex, bin, oct literal
        # TODO: remove width inference from text
        else:
            value = int(text, 0)
            if text.startswith("0b"):
                width = len(text) - 2
            elif text.startswith("0x"):
                width = (len(text) - 2) * 4
            elif text.startswith("0") and len(text) > 1:
                width = (len(text) - 1) * 3
            else:
                width = value.bit_length()

        return behav.IntLiteral(value, width)

    def visitDeclaration(self, ctx: CoreDSL2Parser.DeclarationContext):
        """Generate a declaration."""

        # extract storage type, qualifiers and attributes
        storage = [self.visit(obj) for obj in ctx.storage]
        # qualifiers = [self.visit(obj) for obj in ctx.qualifiers]
        attributes = dict([self.visit(obj) for obj in ctx.attributes])

        # extract data type
        type_ = self.visit(ctx.type_)

        # extract list of contained declarations for the given type
        decls: "list[CoreDSL2Parser.DeclaratorContext]" = ctx.declarations

        ret_decls = []

        # generate each declaration
        for decl in decls:
            name = decl.name.text

            # generate a register alias
            if type_.ptr == "&":
                # error out on duplicate declaration
                if name in self._memory_aliases:
                    raise M2DuplicateError(f"memory {name} already defined")

                # assume default size
                size = [1]
                # alias needs to have a reference as initializer
                init: behav.IndexedReference = self.visit(decl.init)
                attributes = {}

                # extract array size
                if decl.size:
                    size = [self.visit(obj).value for obj in decl.size]

                # extract referenced object and indices
                left = init.index
                right = init.right if init.right is not None else left
                reference = init.reference

                if decl.attributes:
                    attributes = dict([self.visit(obj) for obj in decl.attributes])

                range = arch.RangeSpec(left, right)

                # if range.length != size[0]:
                # 	raise ValueError(f"range mismatch for \"{name}\"")

                # instantiate M2-ISA-R object, keep track of parent - child relations
                m = arch.Memory(name, range, type_._width, attributes)
                m.parent = reference
                m.parent.children.append(m)

                # keep track of this declaration globally
                self._memory_aliases[name] = m
                # keep track of this declaration for this declaration statement
                ret_decls.append(m)

            # normal declaration
            else:
                # no storage specifier -> implementation parameter, "Constant" in M2-ISA-R
                if len(storage) == 0:
                    if name in self._constants:
                        raise M2DuplicateError(f"constant {name} already defined")

                    # extract initializer if present
                    init = None
                    if decl.init is not None:
                        init = self.visit(decl.init)

                    c = arch.Constant(name, init, [], type_._width, type_.signed)

                    self._constants[name] = c
                    ret_decls.append(c)

                # register and extern declaration: "Memory" object in M2-ISA-R
                elif "register" in storage or "extern" in storage:
                    if name in self._memories:
                        raise M2DuplicateError(f"memory {name} already defined")

                    size = [1]
                    init = None
                    attributes = {}

                    if decl.size:
                        size = [self.visit(obj) for obj in decl.size]

                    if len(size) > 1:
                        raise NotImplementedError("arrays with more than one dimension are not supported")

                    if decl.init is not None:
                        init = self.visit(decl.init)

                    if decl.attributes:
                        attributes = dict([self.visit(obj) for obj in decl.attributes])

                    range = arch.RangeSpec(size[0])
                    m = arch.Memory(name, range, type_._width, attributes)

                    # attach init value to memory object
                    if init is not None:
                        m._initval[None] = init.generate(None)

                    if arch.MemoryAttribute.IS_MAIN_REG in attributes:
                        self._main_reg_file = m

                    self._memories[name] = m
                    ret_decls.append(m)

        return ret_decls

    def visitType_specifier(self, ctx: CoreDSL2Parser.Type_specifierContext):
        type_ = self.visit(ctx.type_)
        if ctx.ptr:
            type_.ptr = ctx.ptr.text
        return type_

    def visitInteger_type(self, ctx: CoreDSL2Parser.Integer_typeContext):
        """Generate an integer type specification."""

        # default signedness
        signed = True
        # minimal integer type is just a signedness without width
        width = None

        # extract sign
        if ctx.signed is not None:
            signed = self.visit(ctx.signed)

        # extract size
        if ctx.size is not None:
            width = self.visit(ctx.size)

        # extract and decode shorthand (int = signed<32>)
        if ctx.shorthand is not None:
            width = self.visit(ctx.shorthand)

        # type check width
        if isinstance(width, behav.IntLiteral):
            width = width.value
        elif isinstance(width, behav.NamedReference):
            width = width.reference
        else:
            raise M2TypeError("width has wrong type")

        return arch.IntegerType(width, signed, None)

    def visitVoid_type(self, ctx: CoreDSL2Parser.Void_typeContext):
        """Generate a void type."""
        return arch.VoidType(None)

    def visitBool_type(self, ctx: CoreDSL2Parser.Bool_typeContext):
        """Generate a bool (alias for unsigned<1>)."""
        return arch.IntegerType(1, False, None)

    def visitBinary_expression(self, ctx: CoreDSL2Parser.Binary_expressionContext):
        """Generate a binary expression."""

        # visit LHS and RHS
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        op = behav.Operator(ctx.bop.text)

        # return M2-ISA-R object
        return behav.BinaryOperation(left, op, right)

    def visitSlice_expression(self, ctx: CoreDSL2Parser.Slice_expressionContext):
        left = self.visit(ctx.left)
        right = self.visit(ctx.right) if ctx.right is not None else None
        expr = self.visit(ctx.expr).reference

        op = behav.IndexedReference(expr, left, right)
        return op

    def visitPrefix_expression(self, ctx: CoreDSL2Parser.Prefix_expressionContext):
        prefix = behav.Operator(ctx.prefix.text)
        expr = self.visit(ctx.right)
        return behav.UnaryOperation(prefix, expr)

    def visitReference_expression(self, ctx: CoreDSL2Parser.Reference_expressionContext):
        """Generate a referencing expression."""

        name = ctx.ref.text

        # try to resolve the reference, error out if invalid
        ref = self._constants.get(name) or self._memories.get(name) or self._memory_aliases.get(name)
        if ref is None:
            raise M2NameError(f'reference "{name}" could not be resolved')
        return behav.NamedReference(ref)

    def visitStorage_class_specifier(self, ctx: CoreDSL2Parser.Storage_class_specifierContext):
        return ctx.children[0].symbol.text

    def visitType_qualifier(self, ctx: CoreDSL2Parser.Type_qualifierContext):
        return ctx.children[0].symbol.text

    def visitInteger_signedness(self, ctx: CoreDSL2Parser.Integer_signednessContext):
        return SIGNEDNESS[ctx.children[0].symbol.text]

    def visitInteger_shorthand(self, ctx: CoreDSL2Parser.Integer_shorthandContext):
        return behav.IntLiteral(SHORTHANDS[ctx.children[0].symbol.text])

    def visitAssignment_expression(self, ctx: CoreDSL2Parser.Assignment_expressionContext):
        """Generate an assignment."""

        # extract LHS and RHS
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)

        # if LHS is a reference, assign RHS as its default value
        if isinstance(left, behav.NamedReference):
            if isinstance(left.reference, arch.Constant):
                left.reference.value = right.generate(None)

            elif isinstance(left.reference, arch.Memory):
                left.reference._initval[None] = right.generate(None)

        elif isinstance(left, behav.IndexedReference):
            left.reference._initval[left.index.generate(None)] = right.generate(None)

    def visitAttribute(self, ctx: CoreDSL2Parser.AttributeContext):
        """Generate an attribute."""

        name = ctx.name.text

        # read attribute from enums
        attr = (
            arch.InstrAttribute._member_map_.get(name.upper())
            or arch.MemoryAttribute._member_map_.get(name.upper())
            or arch.FunctionAttribute._member_map_.get(name.upper())
        )

        # warn if attribute is unknown to M2-ISA-R
        if attr is None:
            logger.warning('unknown attribute "%s" encountered', name)
            attr = name

        return attr, ctx.params

    def visitChildren(self, node):
        """Helper method to return flatter results on tree visits."""
        # print("visitChildren", node)

        ret = super().visitChildren(node)
        if isinstance(ret, list) and len(ret) == 1:
            return ret[0]
        return ret

    def aggregateResult(self, aggregate, nextResult):
        """Aggregate results from multiple children into a list."""

        ret = aggregate
        if nextResult is not None:
            if ret is None:
                ret = [nextResult]
            else:
                ret += [nextResult]
        return ret



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import os
import argparse
import logging
import pathlib

import pandas as pd

from seal5.tools import cdsl2llvm
from seal5.model import Seal5InstrAttribute
from seal5.riscv_utils import build_riscv_mattr, get_riscv_defaults
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.llvmir_behavior_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set and instruction")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--ext", type=str, default="ll", help="Default file extension (if using --splitted)")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)
    model_name = top_level.stem

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    # print("model", model)
    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }
    settings = model_obj.settings
    if args.splitted:
        # errs = []
        # model_includes = []
        # errs = []
        if settings:
            riscv_settings = settings.riscv
            model_settings = settings.models.get(model_name)
            model_riscv_settings = model_settings.riscv
            if model_riscv_settings is not None:
                riscv_settings = riscv_settings.merge(model_riscv_settings)
        else:
            riscv_settings = None

        assert out_path.is_dir(), "Expecting output directory when using --splitted"
        for set_name, set_def in model_obj.sets.items():
            if len(set_def.instructions) == 0:
                continue
            xlen = set_def.xlen
            metrics["n_sets"] += 1
            ext_settings = set_def.settings
            riscv_settings_ = riscv_settings
            ext_riscv_settings = ext_settings.riscv
            if ext_riscv_settings is not None:
                riscv_settings_ = riscv_settings_.merge(ext_riscv_settings)
            default_features, default_xlen = get_riscv_defaults(riscv_settings)
            if xlen is None:  # TODO: redundant?
                xlen = default_xlen
            for instr_def in set_def.instructions.values():
                metrics["n_instructions"] += 1
                attrs = instr_def.attributes
                if len(attrs) > 0:
                    skip = Seal5InstrAttribute.SKIP_PATTERN_GEN in attrs
                    if skip:
                        metrics["n_skipped"] += 1
                        metrics["skipped_instructions"].append(instr_def.name)
                        continue
                input_file = out_path / set_name / f"{instr_def.name}.core_desc"
                attrs = instr_def.attributes
                skip = False
                if len(attrs) > 0:
                    skip = Seal5InstrAttribute.SKIP_PATTERN_GEN in attrs
                if not input_file.is_file():
                    skip = True
                if skip:
                    metrics["n_skipped"] += 1
                    metrics["skipped_instructions"].append(instr_def.name)
                    # errs.append(TODO)
                output_file = out_path / set_name / f"{instr_def.name}.{args.ext}"

                features = [*default_features]
                if ext_settings is not None:
                    arch_ = ext_settings.get_arch(name=set_name)
                    if arch_ is not None:
                        features.append(arch_)
                    riscv_settings = ext_settings.riscv
                    if riscv_settings is not None:
                        xlen_ = riscv_settings.xlen
                        if xlen_ is not None:
                            xlen = xlen_
                mattr = build_riscv_mattr(features, xlen)

                install_dir = os.getenv("CDSL2LLVM_DIR", None)
                assert install_dir is not None
                install_dir = pathlib.Path(install_dir)
                try:
                    cdsl2llvm.run_pattern_gen(
                        # install_dir / "llvm" / "build",
                        install_dir,
                        input_file,
                        output_file,
                        skip_patterns=True,
                        skip_formats=True,
                        mattr=mattr,
                        xlen=xlen,
                    )
                    metrics["n_success"] += 1
                    metrics["success_instructions"].append(instr_def.name)
                except AssertionError:
                    metrics["n_failed"] += 1
                    metrics["failed_instructions"].append(instr_def.name)
                    # errs.append((insn_name, str(ex)))
        # if len(errs) > 0:
        #     # print("errs", errs)
        #     for insn_name, err_str in errs:
        #         print("Err:", insn_name, err_str)
        #         input("!")
        #         cdsl2llvm.
        #         out_path_ = out_path / set_name / f"{instr_def.name}.{args.ext}"
        #         out_path_.parent.mkdir(exist_ok=True)
    else:
        raise NotImplementedError
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Common Seal5 CLI code."""
from seal5.logging import Logger

logger = Logger("cli")


def handle_logging_flags(args):
    """Handler for logging-related flags."""
    logger.setLevel(args.log.upper())


def add_common_options(parser):
    """Defines options used by multiple cmdline parsers."""
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print tool outputs for easier debugging (default: %(default)s)",
    )
    parser.add_argument(
        "-n",
        "--name",
        metavar="NAME",
        type=str,
        default=None,
        help="Environment name",
    )
    parser.add_argument(
        "--dir",
        type=str,
        default=None,
        help="LLVM directory (default: %(default)s",
    )



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for Applying Seal5 patches."""

from seal5.flow import Seal5Flow
from seal5.logging import Logger
from seal5.types import PatchStage


logger = Logger("cli")


def add_patch_options(parser):
    """Setup parser for patch argument group."""
    patch_parser = parser.add_argument_group("patch options")
    patch_parser.add_argument(
        "--stages",
        "-s",
        nargs="+",
        type=str,
        default=1,
        help="Patched Seal5 LLVM after stages",
    )
    patch_parser.add_argument("--force", "-f", default=False, action="store_true", help="Force patch in LLVM")


def get_parser(subparsers):
    """ "Define and return a subparser for the patch subcommand."""
    parser = subparsers.add_parser("patch", description="Applying Seal5 patches")
    parser.set_defaults(func=handle)
    add_patch_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the patch subcommand"""
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.patch(verbose=args.verbose, stages=[PatchStage(int(i)) for i in args.stages], force=args.force)



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for cleaning seal5 environment."""

from seal5.flow import Seal5Flow
from seal5.logging import Logger


logger = Logger("cli")


def add_clean_options(parser):
    """Setup parser for clean argument group."""
    clean_parser = parser.add_argument_group("clean options")
    clean_parser.add_argument(
        "--temp",
        action="store_true",
        help="Delete temp folder folder?",
    )
    clean_parser.add_argument(
        "--patches",
        action="store_true",
        help="Delete patches folder folder?",
    )
    clean_parser.add_argument(
        "--models",
        action="store_true",
        help="Delete models folder folder?",
    )
    clean_parser.add_argument(
        "--inputs",
        action="store_true",
        help="Delete inputs folder folder?",
    )
    clean_parser.add_argument(
        "--logs",
        action="store_true",
        help="Delete logs folder folder?",
    )
    clean_parser.add_argument(
        "--install",
        action="store_true",
        help="Delete install folder folder?",
    )
    clean_parser.add_argument(
        "--build",
        action="store_true",
        help="Delete build folder folder?",
    )
    clean_parser.add_argument(
        "--deps",
        action="store_true",
        help="Delete deps folder folder?",
    )
    clean_parser.add_argument(
        "--non-interactive",
        default=True,
        action="store_true",
        help="Do not ask questions interactively",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the clean subcommand."""
    parser = subparsers.add_parser("clean", description="clean Seal5 settings.")
    parser.set_defaults(func=handle)
    add_clean_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the clean subcommand"""
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.clean(
        temp=args.temp,
        patches=args.patches,
        models=args.models,
        inputs=args.inputs,
        logs=args.logs,
        install=args.install,
        build=args.build,
        deps=args.deps,
        verbose=args.verbose,
        interactive=not args.non_interactive,
    )



from typing import Optional

from seal5.settings import RISCVSettings
from seal5.model import Seal5OperandAttribute

from m2isar.metamodel import arch


def get_riscv_defaults(riscv_settings: Optional[RISCVSettings] = None):
    default_features = ["+m", "+fast-unaligned-access"]
    default_xlen = None
    if riscv_settings:
        default_features_ = riscv_settings.features
        if default_features_ is not None:
            default_features = default_features_
        default_xlen_ = riscv_settings.xlen
        if default_xlen is not None:
            default_xlen = default_xlen_
    return default_features, default_xlen


def build_riscv_mattr(features, xlen: Optional[int] = None):
    mattrs = []

    def fix_prefix(x):
        assert len(x) > 0
        if x[0] in ["-", "+"]:
            return x
        return f"+{x}"

    if features is not None:
        mattrs += list(features)
    if xlen is not None:
        if xlen == 64:
            mattrs += ["64bit"]
    mattrs = list(set(map(fix_prefix, mattrs)))
    mattr = ",".join(mattrs)
    return mattr


def detect_opcode(instr_def):  # TODO: move to transform and store as attr
    OPCODE_LOOKUP = {
        "LOAD": 0b00000,
        "LOAD-FP": 0b00001,
        "custom-0": 0b00010,
        "MISC-MEM": 0b00011,
        "OP-IMM": 0b00100,
        "AUIPC": 0b00101,
        "OP-IMM-32": 0b00110,
        # "48bit": 0b00111,
        "STORE": 0b01000,
        "STORE-FP": 0b01001,
        "custom-1": 0b01010,
        "AMO": 0b01011,
        "OP": 0b01100,
        "LUI": 0b01101,
        "OP-32": 0b01110,
        # "64bit": 0b01111,
        "MADD": 0b10000,
        "MSUB": 0b10001,
        "NMADD": 0b10010,
        "NMSUB": 0b10011,
        "OP-FP": 0b10100,
        "OP-V": 0b10101,
        "custom-2": 0b10110,  # rv128i
        # "48bit2": 0b10111,
        "BRANCH": 0b11000,
        "JALR": 0b11001,
        # "reserved": 0b11010,
        "JAL": 0b11011,
        "SYSTEM": 0b11100,
        "OP-P": 0b11101,
        "custom-3": 0b11110,
        # "80bit+": 0b11111,
    }
    OPCODE_LOOKUP_REV = {v: k for k, v in OPCODE_LOOKUP.items()}
    size = 0  # TODO: use instr_def.size
    opcode = None
    for e in reversed(instr_def.encoding):
        # print("e", e, dir(e))
        if isinstance(e, arch.BitVal):
            length = e.length
            if size == 0:
                val = e.value
                if length >= 7:
                    val = val & 0b1111111
                    opcode = val >> 2
        elif isinstance(e, arch.BitField):
            length = e.range.length
        else:
            assert False
        size += length
    if size != 32:  # TODO: support compressed opcodes
        assert size == 16
        return "COMPRESSED", opcode
    assert opcode is not None
    found = OPCODE_LOOKUP_REV.get(opcode, None)
    assert found is not None, f"Opcode not found: {bin(opcode)}"
    return found, opcode


def detect_funct3_funct7(instr_def):
    size = 0  # TODO: use instr_def.size
    funct3 = None
    funct7 = None
    if size != 32:  # TODO: support compressed opcodes
        return None, None
    for e in reversed(instr_def.encoding):
        # print("e", e, dir(e))
        if isinstance(e, arch.BitVal):
            length = e.length
            if size == 7 + 5:
                val = e.value
                if length == 3:
                    funct3 = val
            elif size == 7 + 5 + 3 + 5 + 5:
                val = e.value
                if length == 7:
                    funct7 = val
        elif isinstance(e, arch.BitField):
            length = e.range.length
        else:
            assert False
        size += length
        assert size == 16
        return "COMPRESSED"
    return funct3, funct7


def detect_format(instr_def):  # TODO: move to transform and store as attr
    # enc = instr_def.encoding
    operands = instr_def.operands
    char_lookup = {}
    imm_char = "a"
    reg_char = "A"
    temp = ""
    for e in reversed(instr_def.encoding):
        if isinstance(e, arch.BitVal):
            new = bin(e.value)[2:].zfill(e.length)
            temp = f"{new}{temp}"
        elif isinstance(e, arch.BitField):
            length = e.range.length
            found = char_lookup.get(e.name, None)
            char = "#"
            if found is not None:
                char = found
            else:
                op = operands.get(e.name, None)
                if op is not None:
                    if Seal5OperandAttribute.IS_REG in op.attributes:
                        char = reg_char
                        reg_char = chr(ord(reg_char) + 1)
                    elif Seal5OperandAttribute.IS_IMM in op.attributes:
                        char = imm_char
                        imm_char = chr(ord(imm_char) + 1)
                    else:
                        char = "%"
                char_lookup[e.name] = char
            new = char * length
            temp = f"{new}{temp}"
        else:
            assert False
    FMT_LOOKUP = [  # needs to be sorted by weight (number of non-?)
        ("aaaaaaaBBBBBAAAAA???aaaaa???????", "s-type (mem)"),
        ("??aaaaaCCCCCBBBBB???AAAAA???????", "f2-type (binop)"),
        ("??bbbbbaaaaaBBBBB???AAAAA???????", "f2-type (unop)"),
        ("???????CCCCCBBBBB???AAAAA???????", "r-type (binop)"),
        ("???????BBBBBAAAAA???00000???????", "r-type (binop, noout)"),
        ("???????00000BBBBB???AAAAA???????", "r-type (unop)"),
        ("???????00000BBBBB???00000???????", "r-type (unop, noout)"),
        ("???????0000000000???AAAAA???????", "r-type (nonop)"),
        ("???????0000000000???00000???????", "r-type (nonop)"),  # s4emac
        ("???????aaaaaBBBBB???AAAAA???????", "ri-type (binop)"),
    ]
    found2 = f"unknown[{temp}]"
    for mask, fmt in FMT_LOOKUP:
        if len(mask) != len(temp):
            continue
        masked = "".join([c if mask[i] != "?" else "?" for i, c in enumerate(temp)])
        if masked == mask:
            found2 = fmt
            break
    return found2, temp



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .transform import main as InferTypes

__all__ = ["InferTypes"]



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""A transformation module for simplifying M2-ISA-R behavior expressions. The following
simplifications are done:

* Resolvable :class:`m2isar.metamodel.arch.Constant` s are replaced by
  `m2isar.metamodel.arch.IntLiteral` s representing their value
* Fully resolvable arithmetic operations are carried out and their results
  represented as a matching :class:`m2isar.metamodel.arch.IntLiteral`
* Conditions and loops with fully resolvable conditions are either discarded entirely
  or transformed into code blocks without any conditions
* Ternaries with fully resolvable conditions are transformed into only the matching part
* Type conversions of :class:`m2isar.metamodel.arch.IntLiteral` s apply the desired
  type directly to the :class:`IntLiteral` and discard the type conversion
"""
from copy import copy

from m2isar.metamodel import arch, behav

from seal5.logging import Logger

logger = Logger("transform." + __name__)

# pylint: disable=unused-argument


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        # try:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            statements.extend(temp)
        else:
            statements.append(temp)
        # except (NotImplementedError, ValueError):
        #   print(f"cant simplify {stmt}")

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    # print("binary_operation")
    # print("self.op.value", self.op.value)

    self.left = self.left.generate(context)
    self.right = self.right.generate(context)
    # print("self.left", self.left)
    # print("self.right", self.right)

    # print("self.left.inferred_type", self.left.inferred_type)
    # print("self.right.inferred_type", self.right.inferred_type)

    # see: https://github.com/Minres/CoreDSL/wiki/Expressions#arithmetic-type-rules
    if self.op.value in ["+", "-", "*", "/", "%", "|", "&", "^", "<<", ">>"]:
        if self.left.inferred_type is None or self.right.inferred_type is None:
            logger.warning("Slice Operation needs inferred type. Skipping...")
            self.inferred_type = None
            return self
        assert isinstance(self.left.inferred_type, arch.IntegerType)
        assert isinstance(self.right.inferred_type, arch.IntegerType)
        w1 = self.left.inferred_type._width
        w2 = self.right.inferred_type._width
        s1 = self.left.inferred_type.signed
        s2 = self.right.inferred_type.signed
        if self.op.value == "+":
            if not s1 and not s2:
                wr = max(w1, w2) + 1
                sr = False
            elif s1 and s2:
                wr = max(w1, w2) + 1
                sr = True
            elif s1 and not s2:
                wr = max(w1, w2 + 1) + 1
                sr = True
            elif not s1 and s2:
                wr = max(w1 + 1, w2) + 1
                sr = True
        elif self.op.value == "-":
            sr = True
            if not s1 and not s2:
                wr = max(w1 + 1, w2 + 1)
            elif s1 and s2:
                wr = max(w1 + 1, w2 + 1)
            elif s1 and not s2:
                wr = max(w1, w2 + 1) + 1
            elif not s1 and s2:
                wr = max(w1 + 1, w2) + 1
        elif self.op.value == "*":
            wr = w1 + w2
            sr = s1 or s2
        elif self.op.value == "/":
            wr = w1 if not s2 else (w1 + 1)
            sr = s1 or s2
        elif self.op.value == "%":
            if not s1 and not s2:
                wr = min(w1, w2)
                sr = False
            elif s1 and s2:
                wr = min(w1, w2)
                sr = True
            elif s1 and not s2:
                wr = min(w1, w2 + 1)
                sr = True
            elif not s1 and s2:
                wr = min(w1, max(1, w2 - 1))
                sr = False
        elif self.op.value in ["|", "&", "^"]:
            wr = max(w1, w2)
            sr = s1 or s2
        elif self.op.value in [">>", "<<"]:
            wr = w1
            sr = s1
        self.inferred_type = arch.IntegerType(wr, sr, None)
    else:
        if self.op.value in ["||", "&&"]:
            self.inferred_type = arch.IntegerType(1, False, None)  # unsigned<1> / bool
        elif self.op.value in ["<", ">", "==", "!=", ">=", "<="]:
            self.inferred_type = arch.IntegerType(1, False, None)  # unsigned<1> / bool
    # print("sit", self.inferred_type.width)
    assert self.inferred_type is not None
    # input("!x!")

    return self


def slice_operation(self: behav.SliceOperation, context):
    # print("slice_operation")
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)
    # print("self.expr", self.expr)
    # print("self.left", self.left)
    # print("self.right", self.right)
    # input("slice")

    # type inference
    if self.expr.inferred_type is None:
        logger.warning("Slice Operation needs inferred type. Skipping...")
        return self
    assert isinstance(self.expr.inferred_type, arch.IntegerType)
    ty = self.expr.inferred_type
    # For non-static slices, we cann not infer the type!
    if not isinstance(self.left, behav.IntLiteral):
        logger.warning("Can not infer type of non-static slice operation. Skipping...")
        return self
    lval = self.left.value
    if not isinstance(self.right, behav.IntLiteral):
        logger.warning("Can not infer type of non-static slice operation. Skipping...")
        return self
    rval = self.right.value
    width = lval - rval + 1 if lval > rval else rval - lval + 1
    ty_ = copy(ty)
    ty_._width = width
    self.inferred_type = ty_
    # print("sit", ty)
    # input("*")

    return self


def concat_operation(self: behav.ConcatOperation, context):
    # print("concat_coperation")
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    # print("number_literal")
    if isinstance(self, behav.IntLiteral):
        bit_size = self.bit_size
        signed = self.signed

        self.inferred_type = arch.IntegerType(bit_size, signed, None)
    return self


def int_literal(self: behav.IntLiteral, context):
    # print("int_literal")

    # type inference
    bit_size = self.bit_size
    signed = self.signed

    self.inferred_type = arch.IntegerType(bit_size, signed, None)

    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    # type inference
    # print("scalar_definition", self, dir(self), self.scalar, self.scalar.size, self.scalar.data_type)
    signed = self.scalar.data_type == arch.DataType.S
    width = self.scalar.size
    self.inferred_type = arch.IntegerType(width, signed, None)
    return self


def assignment(self: behav.Assignment, context):
    # print("assignment", self)
    # print("at_", self.target)
    # print("ae_", self.expr)
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    # if isinstance(self.expr, behav.IntLiteral) and isinstance(self.target, behav.ScalarDefinition):
    #       self.target.scalar.value = self.expr.value

    # type inference
    # print("at", self.target)
    # print("ae", self.expr)
    # print("at1", self.target.inferred_type)
    # print("ae1", self.expr.inferred_type)
    # print("at2", self.target.inferred_type.width)
    # print("ae2", self.expr.inferred_type.width)
    # input("ccc")
    self.inferred_type = None

    return self


def conditional(self: behav.Conditional, context):
    self.conds = [x.generate(context) for x in self.conds]
    # self.stmts = [[y.generate(context) for y in x] for x in self.stmts]
    stmts = []
    for stmt in self.stmts:
        if isinstance(stmt, list):  # TODO: legacy?
            new = [y.generate(context) for y in stmt]
        else:
            new = stmt.generate(context)
        stmts.append(new)
    self.stmts = stmts

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    # print("ternary")

    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    # print("ste", self.then_expr)
    # print("see", self.else_expr)
    # print("ste1", self.then_expr.inferred_type)
    # print("see1", self.else_expr.inferred_type)
    # print("ste2", self.then_expr.inferred_type.width)
    # print("see2", self.else_expr.inferred_type.width)
    # TODO
    then_ty = self.then_expr.inferred_type
    else_ty = self.else_expr.inferred_type
    if then_ty and else_ty:
        # assert then_ty.signed == else_ty.signed
        wt = then_ty.width
        we = else_ty.width
        wr = max(wt, we)
        # print("wr", wr)
        # input("o")
        self.inferred_type = arch.IntegerType(wr, True, None)
    # input("ppp")

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    # print("unary_operation")

    self.right = self.right.generate(context)

    # print("sr", self.right)
    # print("sr1", self.right.inferred_type)
    # print("sr2", self.right.inferred_type.width)
    # input("!")
    if self.right.inferred_type:
        w1 = self.right.inferred_type.width
        if self.op.value == "-":
            inferred_type = arch.IntegerType(w1 + 1, True, None)
        elif self.op.value == "~":
            inferred_type = arch.IntegerType(w1, True, None)
        elif self.op.value == "!":
            inferred_type = arch.IntegerType(1, False, None)
        else:
            inferred_type = None
        self.inferred_type = inferred_type

    return self


def named_reference(self: behav.NamedReference, context):
    # print("named_reference", self)
    # print("dir", dir(self))
    # print("self.reference", self.reference)
    reference = self.reference

    # type inference
    # self.infered_type = ?
    if isinstance(reference, arch.BitFieldDescr):
        # print("BITFIELD")
        # print("self.reference", self.reference)
        # print("self.reference.data_type", self.reference.data_type)
        assert self.reference.data_type in [arch.DataType.U, arch.DataType.S]
        ty = arch.IntegerType(reference.size, reference.data_type == arch.DataType.S, None)
        # print("ty", ty)
        self.inferred_type = ty

    elif isinstance(reference, arch.Scalar):
        dt = reference.data_type
        sz = reference.size
        assert dt in [arch.DataType.U, arch.DataType.S]
        signed = dt == arch.DataType.S
        ty = arch.IntegerType(sz, signed, None)
        self.inferred_type = ty
    elif isinstance(reference, arch.Memory):
        if reference.is_pc:  # TODO: check if special case required?
            self.inferred_type = arch.IntegerType(reference.size, False, None)
        # print("type(reference)", type(reference))

    # print("self.inferred_type", self.inferred_type)
    # input("222")

    return self


def indexed_reference(self: behav.IndexedReference, context):
    # print("indexed_reference")
    self.index = self.index.generate(context)

    # type inference
    assert isinstance(self.reference, arch.Memory)
    ty = arch.DataType.U  # TODO: Memory class should keep track of dtype, not only size?
    assert ty in [arch.DataType.U, arch.DataType.S]
    size = self.reference.size
    ty_ = arch.IntegerType(size, ty == arch.DataType.S, None)

    self.inferred_type = ty_
    # print("self.inferred_type", self.inferred_type)
    # input("111")

    return self


def type_conv(self: behav.TypeConv, context):
    # print("type_conv")
    self.expr = self.expr.generate(context)
    # print("self.expr", self.expr)

    ty = self.expr.inferred_type
    if ty is None:
        logger.warning("Type conv needs inferred type. Skipping...")
        return self
    assert isinstance(ty, arch.IntegerType)
    assert self.data_type in [arch.DataType.U, arch.DataType.S]
    ty.signed = self.data_type == arch.DataType.S
    if self.size is not None:
        ty._width = self.size

    # type inference
    self.inferred_type = ty

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def procedure_call(self: behav.ProcedureCall, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    if isinstance(self.expr, behav.IntLiteral):
        return self.expr

    # type inference
    self.inferred_type = self.expr.inferred_type

    return self


def break_(self: behav.Break, context):
    return self



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Diff report writer for Seal5."""

import argparse
import logging
import pathlib

import pandas as pd

from seal5.settings import Seal5Settings

from seal5.logging import Logger

logger = Logger("backends.diff_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md", "mermaid"], default="auto")
    parser.add_argument("--yaml", type=str, default=None)
    # TODO: extract info per patch/model
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    assert args.output is not None
    out_path = pathlib.Path(args.output)
    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    def process_metrics(settings, ignore_passes: bool = False):
        metrics = settings.metrics
        assert metrics is not None
        assert isinstance(metrics, list)
        ret = []
        for stage_metrics in metrics:
            assert isinstance(stage_metrics, dict)
            assert len(stage_metrics) == 1
            stage_name = list(stage_metrics.keys())[0]
            stage_metrics = list(stage_metrics.values())[0]
            assert isinstance(stage_metrics, dict)
            if stage_name == "deploy":
                n_files_changed = stage_metrics.get("n_files_changed", None)
                n_insertions = stage_metrics.get("n_insertions", None)
                n_deletions = stage_metrics.get("n_deletions", None)
                new = {
                    "phase": "*",
                    "n_files_changed": n_files_changed,
                    "n_insertions": n_insertions,
                    "n_deletions": n_deletions,
                }
                ret.append(new)
                continue
            if stage_name != "patch":
                continue
            phases = stage_metrics.get("stages", None)
            if phases is None:
                continue
            assert isinstance(phases, dict)
            for phase_name, phase_metrics in phases.items():
                n_files_changed = phase_metrics.get("n_files_changed", None)
                n_insertions = phase_metrics.get("n_insertions", None)
                n_deletions = phase_metrics.get("n_deletions", None)
                new = {
                    "phase": phase_name,
                    "n_files_changed": n_files_changed,
                    "n_insertions": n_insertions,
                    "n_deletions": n_deletions,
                }
                ret.append(new)
        return ret

    diff_data = process_metrics(settings)

    diff_df = pd.DataFrame(diff_data)

    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        diff_df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        diff_df.to_pickle(out_path)
    elif fmt == "md":
        # diff_df.to_markdown(out_path, tablefmt="grid", index=False)
        diff_df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Loging utilities for Seal5."""

import logging
import logging.handlers
import sys
import socket
import os
import socketserver
import struct
import pickle
from typing import List, Optional
import threading
from seal5.settings import FileLoggingSettings

PROJECT_NAME = "seal5"
HOSTNAME = "localhost"
SEAL5_LOGGING_PORT = int(os.getenv("SEAL5_LOGGING_PORT", 0))
_env_val: Optional[str] = os.getenv("SEAL5_INTERNALS_LOGGING_PORT")
SEAL5_INTERNALS_LOGGING_PORT: Optional[int] = int(_env_val) if _env_val is not None else None
_log_server = None
_logger = None


def resolve_log_level(value: str | int):
    return getattr(logging, value.upper()) if isinstance(value, str) else value


def get_formatter(minimal=False):
    """Returns a log formatter for one on two predefined formats."""
    if minimal:
        fmt = "[%(name)s::%(levelname)s] %(message)s"
    else:
        fmt = "%(asctime)s [%(name)s::%(levelname)s] (%(pathname)s::%(lineno)d %(message)s"
    formatter = logging.Formatter(fmt)
    return formatter


def get_logger(loggername: None | str = None, level=logging.DEBUG):

    server_reachable = False
    if SEAL5_INTERNALS_LOGGING_PORT is not None:
        try:
            with socket.create_connection((HOSTNAME, SEAL5_INTERNALS_LOGGING_PORT), timeout=0.2):
                server_reachable = True
        except OSError:
            server_reachable = False

    # --- Case 1: Logging server not reachable → local fallback logger fallback ---
    if not server_reachable:
        fallback_logger = logging.getLogger("fallback")  # fallback logger
        fallback_logger.setLevel(logging.DEBUG)

        # Ensure a StreamHandler exists only once
        if not any(isinstance(h, logging.StreamHandler) for h in fallback_logger.handlers):
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setFormatter(get_formatter(True))
            stream_handler.setLevel(logging.DEBUG)
            fallback_logger.addHandler(stream_handler)

        return fallback_logger

    logger = logging.getLogger(f"{PROJECT_NAME}.{loggername if loggername is not None else 'unknown'}")
    logger.handlers = []
    logger.setLevel(level=level)
    socket_handler = logging.handlers.SocketHandler(HOSTNAME, SEAL5_INTERNALS_LOGGING_PORT)
    logger.addHandler(socket_handler)
    return logger


def initialize_logging_server(
    logfiles: None | List[FileLoggingSettings] = [
        FileLoggingSettings(filename="log_debug.log", level=logging.DEBUG, rotate=True, limit=3),
    ],
    stream_log_level: int | str = logging.INFO,
):
    global _log_server, SEAL5_INTERNALS_LOGGING_PORT, _logger
    logger = logging.getLogger(PROJECT_NAME)
    _logger = logger
    # This should be the lowest value and not changeable since logger is the first filter
    logger.setLevel(logging.DEBUG)

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(get_formatter(True))
    stream_handler.setLevel(resolve_log_level(stream_log_level))
    logger.addHandler(stream_handler)

    if logfiles is not None:
        for config in logfiles:
            if config.rotate:
                file_handler = logging.handlers.RotatingFileHandler(
                    config.filename, maxBytes=1000000000, backupCount=config.limit if config.limit else 3
                )
            else:
                file_handler = logging.FileHandler(config.filename, "w")
            file_handler.setFormatter(get_formatter(False))
            file_handler.setLevel(resolve_log_level(config.level))
            logger.addHandler(file_handler)

    try:
        _log_server = LogRecordSocketReceiver(HOSTNAME, SEAL5_LOGGING_PORT)
        addr = _log_server.server_address
        assert len(addr) == 2
        SEAL5_INTERNALS_LOGGING_PORT = addr[1]
    except OSError as e:
        if e.errno == 98:  # Address already in use
            raise RuntimeError("Initialization of logging Server not possible! Port Taken!")
        else:
            raise  # rethrow unexpected errors

    thread = threading.Thread(target=_log_server.serve_forever)
    thread.daemon = True
    thread.start()
    print(f"Logger started on port {SEAL5_INTERNALS_LOGGING_PORT}")


def stop_logging_server():
    if _log_server is not None:
        print(f"Logger on port {SEAL5_INTERNALS_LOGGING_PORT} stopped.")
        _log_server.shutdown()
        _log_server.server_close()


def check_logging_server():
    if _log_server is not None:
        assert SEAL5_INTERNALS_LOGGING_PORT is not None, "Logging server exists with undef port"
        return True
    return False


def update_rotary_logger():
    if _logger is not None:
        for handler in _logger.handlers[:]:
            if isinstance(handler, logging.handlers.RotatingFileHandler):
                handler.doRollover()


def update_log_level(console_level=None, file_level=None):
    """Set command line or file log level at runtime."""
    if _logger is not None:
        for handler in _logger.handlers[:]:
            if (
                isinstance(handler, (logging.FileHandler, logging.handlers.RotatingFileHandler))
                and file_level is not None
            ):
                file_level = resolve_log_level(file_level)
                handler.setLevel(file_level)
            elif isinstance(handler, logging.StreamHandler) and console_level is not None:
                console_level = resolve_log_level(console_level)
                handler.setLevel(console_level)


# --- Server (listener) that receives LogRecords ---
class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    def handle(self):
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk += self.connection.recv(slen - len(chunk))
            record = logging.makeLogRecord(pickle.loads(chunk))
            logger = logging.getLogger(record.name)
            # The logger is sent, if it holds a socket_handler it will answer ending in a loop
            socket_handlers = [h for h in logger.handlers if isinstance(h, logging.handlers.SocketHandler)]
            for h in socket_handlers:
                logger.removeHandler(h)
            logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    allow_reuse_address = True
    daemon_threads = True

    def __init__(self, host=None, port=None):
        assert host is not None, "Undefined host"
        assert port is not None, "Undefined port"
        super().__init__((host, port), LogRecordStreamHandler)


class Logger:
    """Proxy that initializes its logger only when first used."""

    def __init__(self, name=None):
        self._name = name
        self._logger = None

    def _get_logger(self):
        if self._logger is None:
            self._logger = get_logger(self._name)
        else:
            # replace fallback logger with actual one of server is available
            if self._logger.name == "fallback" and check_logging_server():
                self._logger = get_logger(self._name)
        return self._logger

    def __getattr__(self, attr):
        return getattr(self._get_logger(), attr)
